from .wrappers import MessageEnumTraits, EnumTraits, EnumType, FieldDetails


# CommandFloatField
CommandFloatVelocity = MessageEnumTraits(0, 'CommandFloatVelocity', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad/s', 'velocity', 'Velocity', 'velocity', True))
CommandFloatEffort = MessageEnumTraits(1, 'CommandFloatEffort', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('N*m', 'effort', 'Effort', 'effort', True))
CommandFloatPositionKp = MessageEnumTraits(2, 'CommandFloatPositionKp', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionKp', 'PositionKp', 'position_kp', True))
CommandFloatPositionKi = MessageEnumTraits(3, 'CommandFloatPositionKi', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionKi', 'PositionKi', 'position_ki', True))
CommandFloatPositionKd = MessageEnumTraits(4, 'CommandFloatPositionKd', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionKd', 'PositionKd', 'position_kd', True))
CommandFloatPositionFeedForward = MessageEnumTraits(5, 'CommandFloatPositionFeedForward', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionFeedForward', 'PositionFeedForward', 'position_feed_forward', True))
CommandFloatPositionDeadZone = MessageEnumTraits(6, 'CommandFloatPositionDeadZone', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionDeadZone', 'PositionDeadZone', 'position_dead_zone', True))
CommandFloatPositionIClamp = MessageEnumTraits(7, 'CommandFloatPositionIClamp', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionIClamp', 'PositionIClamp', 'position_i_clamp', True))
CommandFloatPositionPunch = MessageEnumTraits(8, 'CommandFloatPositionPunch', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionPunch', 'PositionPunch', 'position_punch', True))
CommandFloatPositionMinTarget = MessageEnumTraits(9, 'CommandFloatPositionMinTarget', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionMinTarget', 'PositionMinTarget', 'position_min_target', True))
CommandFloatPositionMaxTarget = MessageEnumTraits(10, 'CommandFloatPositionMaxTarget', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionMaxTarget', 'PositionMaxTarget', 'position_max_target', True))
CommandFloatPositionTargetLowpass = MessageEnumTraits(11, 'CommandFloatPositionTargetLowpass', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionTargetLowpass', 'PositionTargetLowpass', 'position_target_lowpass', True))
CommandFloatPositionMinOutput = MessageEnumTraits(12, 'CommandFloatPositionMinOutput', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionMinOutput', 'PositionMinOutput', 'position_min_output', True))
CommandFloatPositionMaxOutput = MessageEnumTraits(13, 'CommandFloatPositionMaxOutput', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionMaxOutput', 'PositionMaxOutput', 'position_max_output', True))
CommandFloatPositionOutputLowpass = MessageEnumTraits(14, 'CommandFloatPositionOutputLowpass', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionOutputLowpass', 'PositionOutputLowpass', 'position_output_lowpass', True))
CommandFloatVelocityKp = MessageEnumTraits(15, 'CommandFloatVelocityKp', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityKp', 'VelocityKp', 'velocity_kp', True))
CommandFloatVelocityKi = MessageEnumTraits(16, 'CommandFloatVelocityKi', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityKi', 'VelocityKi', 'velocity_ki', True))
CommandFloatVelocityKd = MessageEnumTraits(17, 'CommandFloatVelocityKd', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityKd', 'VelocityKd', 'velocity_kd', True))
CommandFloatVelocityFeedForward = MessageEnumTraits(18, 'CommandFloatVelocityFeedForward', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityFeedForward', 'VelocityFeedForward', 'velocity_feed_forward', True))
CommandFloatVelocityDeadZone = MessageEnumTraits(19, 'CommandFloatVelocityDeadZone', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityDeadZone', 'VelocityDeadZone', 'velocity_dead_zone', True))
CommandFloatVelocityIClamp = MessageEnumTraits(20, 'CommandFloatVelocityIClamp', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityIClamp', 'VelocityIClamp', 'velocity_i_clamp', True))
CommandFloatVelocityPunch = MessageEnumTraits(21, 'CommandFloatVelocityPunch', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityPunch', 'VelocityPunch', 'velocity_punch', True))
CommandFloatVelocityMinTarget = MessageEnumTraits(22, 'CommandFloatVelocityMinTarget', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityMinTarget', 'VelocityMinTarget', 'velocity_min_target', True))
CommandFloatVelocityMaxTarget = MessageEnumTraits(23, 'CommandFloatVelocityMaxTarget', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityMaxTarget', 'VelocityMaxTarget', 'velocity_max_target', True))
CommandFloatVelocityTargetLowpass = MessageEnumTraits(24, 'CommandFloatVelocityTargetLowpass', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityTargetLowpass', 'VelocityTargetLowpass', 'velocity_target_lowpass', True))
CommandFloatVelocityMinOutput = MessageEnumTraits(25, 'CommandFloatVelocityMinOutput', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityMinOutput', 'VelocityMinOutput', 'velocity_min_output', True))
CommandFloatVelocityMaxOutput = MessageEnumTraits(26, 'CommandFloatVelocityMaxOutput', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityMaxOutput', 'VelocityMaxOutput', 'velocity_max_output', True))
CommandFloatVelocityOutputLowpass = MessageEnumTraits(27, 'CommandFloatVelocityOutputLowpass', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityOutputLowpass', 'VelocityOutputLowpass', 'velocity_output_lowpass', True))
CommandFloatEffortKp = MessageEnumTraits(28, 'CommandFloatEffortKp', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortKp', 'EffortKp', 'effort_kp', True))
CommandFloatEffortKi = MessageEnumTraits(29, 'CommandFloatEffortKi', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortKi', 'EffortKi', 'effort_ki', True))
CommandFloatEffortKd = MessageEnumTraits(30, 'CommandFloatEffortKd', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortKd', 'EffortKd', 'effort_kd', True))
CommandFloatEffortFeedForward = MessageEnumTraits(31, 'CommandFloatEffortFeedForward', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortFeedForward', 'EffortFeedForward', 'effort_feed_forward', True))
CommandFloatEffortDeadZone = MessageEnumTraits(32, 'CommandFloatEffortDeadZone', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortDeadZone', 'EffortDeadZone', 'effort_dead_zone', True))
CommandFloatEffortIClamp = MessageEnumTraits(33, 'CommandFloatEffortIClamp', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortIClamp', 'EffortIClamp', 'effort_i_clamp', True))
CommandFloatEffortPunch = MessageEnumTraits(34, 'CommandFloatEffortPunch', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortPunch', 'EffortPunch', 'effort_punch', True))
CommandFloatEffortMinTarget = MessageEnumTraits(35, 'CommandFloatEffortMinTarget', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortMinTarget', 'EffortMinTarget', 'effort_min_target', True))
CommandFloatEffortMaxTarget = MessageEnumTraits(36, 'CommandFloatEffortMaxTarget', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortMaxTarget', 'EffortMaxTarget', 'effort_max_target', True))
CommandFloatEffortTargetLowpass = MessageEnumTraits(37, 'CommandFloatEffortTargetLowpass', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortTargetLowpass', 'EffortTargetLowpass', 'effort_target_lowpass', True))
CommandFloatEffortMinOutput = MessageEnumTraits(38, 'CommandFloatEffortMinOutput', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortMinOutput', 'EffortMinOutput', 'effort_min_output', True))
CommandFloatEffortMaxOutput = MessageEnumTraits(39, 'CommandFloatEffortMaxOutput', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortMaxOutput', 'EffortMaxOutput', 'effort_max_output', True))
CommandFloatEffortOutputLowpass = MessageEnumTraits(40, 'CommandFloatEffortOutputLowpass', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortOutputLowpass', 'EffortOutputLowpass', 'effort_output_lowpass', True))
CommandFloatSpringConstant = MessageEnumTraits(41, 'CommandFloatSpringConstant', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('N/m', 'springConstant', 'SpringConstant', 'spring_constant', True))
CommandFloatReferencePosition = MessageEnumTraits(42, 'CommandFloatReferencePosition', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad', 'referencePosition', 'ReferencePosition', 'reference_position', True))
CommandFloatReferenceEffort = MessageEnumTraits(43, 'CommandFloatReferenceEffort', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('N*m', 'referenceEffort', 'ReferenceEffort', 'reference_effort', True))
CommandFloatVelocityLimitMin = MessageEnumTraits(44, 'CommandFloatVelocityLimitMin', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad/s', 'velocityLimitMin', 'VelocityLimitMin', 'velocity_limit_min', True))
CommandFloatVelocityLimitMax = MessageEnumTraits(45, 'CommandFloatVelocityLimitMax', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad/s', 'velocityLimitMax', 'VelocityLimitMax', 'velocity_limit_max', True))
CommandFloatEffortLimitMin = MessageEnumTraits(46, 'CommandFloatEffortLimitMin', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('N*m', 'effortLimitMin', 'EffortLimitMin', 'effort_limit_min', True))
CommandFloatEffortLimitMax = MessageEnumTraits(47, 'CommandFloatEffortLimitMax', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('N*m', 'effortLimitMax', 'EffortLimitMax', 'effort_limit_max', True))
CommandFloatField = EnumType('CommandFloatField', [ CommandFloatVelocity, CommandFloatEffort, CommandFloatPositionKp, CommandFloatPositionKi, CommandFloatPositionKd, CommandFloatPositionFeedForward, CommandFloatPositionDeadZone, CommandFloatPositionIClamp, CommandFloatPositionPunch, CommandFloatPositionMinTarget, CommandFloatPositionMaxTarget, CommandFloatPositionTargetLowpass, CommandFloatPositionMinOutput, CommandFloatPositionMaxOutput, CommandFloatPositionOutputLowpass, CommandFloatVelocityKp, CommandFloatVelocityKi, CommandFloatVelocityKd, CommandFloatVelocityFeedForward, CommandFloatVelocityDeadZone, CommandFloatVelocityIClamp, CommandFloatVelocityPunch, CommandFloatVelocityMinTarget, CommandFloatVelocityMaxTarget, CommandFloatVelocityTargetLowpass, CommandFloatVelocityMinOutput, CommandFloatVelocityMaxOutput, CommandFloatVelocityOutputLowpass, CommandFloatEffortKp, CommandFloatEffortKi, CommandFloatEffortKd, CommandFloatEffortFeedForward, CommandFloatEffortDeadZone, CommandFloatEffortIClamp, CommandFloatEffortPunch, CommandFloatEffortMinTarget, CommandFloatEffortMaxTarget, CommandFloatEffortTargetLowpass, CommandFloatEffortMinOutput, CommandFloatEffortMaxOutput, CommandFloatEffortOutputLowpass, CommandFloatSpringConstant, CommandFloatReferencePosition, CommandFloatReferenceEffort, CommandFloatVelocityLimitMin, CommandFloatVelocityLimitMax, CommandFloatEffortLimitMin, CommandFloatEffortLimitMax, ])

# CommandHighResAngleField
CommandHighResAnglePosition = MessageEnumTraits(0, 'CommandHighResAnglePosition', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad', 'position', 'Position', 'position', True))
CommandHighResAnglePositionLimitMin = MessageEnumTraits(1, 'CommandHighResAnglePositionLimitMin', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad', 'positionLimitMin', 'PositionLimitMin', 'position_limit_min', True))
CommandHighResAnglePositionLimitMax = MessageEnumTraits(2, 'CommandHighResAnglePositionLimitMax', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad', 'positionLimitMax', 'PositionLimitMax', 'position_limit_max', True))
CommandHighResAngleField = EnumType('CommandHighResAngleField', [ CommandHighResAnglePosition, CommandHighResAnglePositionLimitMin, CommandHighResAnglePositionLimitMax, ])




# CommandEnumField
CommandEnumControlStrategy = MessageEnumTraits(0, 'CommandEnumControlStrategy', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'controlStrategy', 'ControlStrategy', 'control_strategy', True))
CommandEnumMstopStrategy = MessageEnumTraits(1, 'CommandEnumMstopStrategy', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'mstopStrategy', 'MstopStrategy', 'mstop_strategy', True))
CommandEnumMinPositionLimitStrategy = MessageEnumTraits(2, 'CommandEnumMinPositionLimitStrategy', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'minPositionLimitStrategy', 'MinPositionLimitStrategy', 'min_position_limit_strategy', True))
CommandEnumMaxPositionLimitStrategy = MessageEnumTraits(3, 'CommandEnumMaxPositionLimitStrategy', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'maxPositionLimitStrategy', 'MaxPositionLimitStrategy', 'max_position_limit_strategy', True))
CommandEnumField = EnumType('CommandEnumField', [ CommandEnumControlStrategy, CommandEnumMstopStrategy, CommandEnumMinPositionLimitStrategy, CommandEnumMaxPositionLimitStrategy, ])

# CommandBoolField
CommandBoolPositionDOnError = MessageEnumTraits(0, 'CommandBoolPositionDOnError', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionDOnError', 'PositionDOnError', 'position_d_on_error', True))
CommandBoolVelocityDOnError = MessageEnumTraits(1, 'CommandBoolVelocityDOnError', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityDOnError', 'VelocityDOnError', 'velocity_d_on_error', True))
CommandBoolEffortDOnError = MessageEnumTraits(2, 'CommandBoolEffortDOnError', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortDOnError', 'EffortDOnError', 'effort_d_on_error', True))
CommandBoolAccelIncludesGravity = MessageEnumTraits(3, 'CommandBoolAccelIncludesGravity', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'accelIncludesGravity', 'AccelIncludesGravity', 'accel_includes_gravity', True))
CommandBoolField = EnumType('CommandBoolField', [ CommandBoolPositionDOnError, CommandBoolVelocityDOnError, CommandBoolEffortDOnError, CommandBoolAccelIncludesGravity, ])

# CommandNumberedFloatField
CommandNumberedFloatDebug = MessageEnumTraits(0, 'CommandNumberedFloatDebug', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
CommandNumberedFloatField = EnumType('CommandNumberedFloatField', [ CommandNumberedFloatDebug, ])

# CommandIoBankField
CommandIoBankA = MessageEnumTraits(0, 'CommandIoBankA', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
CommandIoBankB = MessageEnumTraits(1, 'CommandIoBankB', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
CommandIoBankC = MessageEnumTraits(2, 'CommandIoBankC', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
CommandIoBankD = MessageEnumTraits(3, 'CommandIoBankD', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
CommandIoBankE = MessageEnumTraits(4, 'CommandIoBankE', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
CommandIoBankF = MessageEnumTraits(5, 'CommandIoBankF', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
CommandIoBankField = EnumType('CommandIoBankField', [ CommandIoBankA, CommandIoBankB, CommandIoBankC, CommandIoBankD, CommandIoBankE, CommandIoBankF, ])

# CommandLedField
CommandLedLed = MessageEnumTraits(0, 'CommandLedLed', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'led', 'Led', 'led', True))
CommandLedField = EnumType('CommandLedField', [ CommandLedLed, ])

# CommandStringField
CommandStringName = MessageEnumTraits(0, 'CommandStringName', allow_broadcast=False, not_bcastable_reason='Cannot set same name for all modules in a group.', yields_dict=None, field_details=None)
CommandStringFamily = MessageEnumTraits(1, 'CommandStringFamily', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
CommandStringAppendLog = MessageEnumTraits(2, 'CommandStringAppendLog', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
CommandStringField = EnumType('CommandStringField', [ CommandStringName, CommandStringFamily, CommandStringAppendLog, ])

# CommandFlagField
CommandFlagSaveCurrentSettings = MessageEnumTraits(0, 'CommandFlagSaveCurrentSettings', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'saveCurrentSettings', 'SaveCurrentSettings', 'save_current_settings', True))
CommandFlagReset = MessageEnumTraits(1, 'CommandFlagReset', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'reset', 'Reset', 'reset', True))
CommandFlagBoot = MessageEnumTraits(2, 'CommandFlagBoot', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'boot', 'Boot', 'boot', True))
CommandFlagStopBoot = MessageEnumTraits(3, 'CommandFlagStopBoot', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'stopBoot', 'StopBoot', 'stop_boot', True))
CommandFlagClearLog = MessageEnumTraits(4, 'CommandFlagClearLog', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'clearLog', 'ClearLog', 'clear_log', True))
CommandFlagField = EnumType('CommandFlagField', [ CommandFlagSaveCurrentSettings, CommandFlagReset, CommandFlagBoot, CommandFlagStopBoot, CommandFlagClearLog, ])
# FeedbackFloatField
FeedbackFloatBoardTemperature = MessageEnumTraits(0, 'FeedbackFloatBoardTemperature', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('C', 'boardTemperature', 'BoardTemperature', 'board_temperature', True))
FeedbackFloatProcessorTemperature = MessageEnumTraits(1, 'FeedbackFloatProcessorTemperature', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('C', 'processorTemperature', 'ProcessorTemperature', 'processor_temperature', True))
FeedbackFloatVoltage = MessageEnumTraits(2, 'FeedbackFloatVoltage', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('V', 'voltage', 'Voltage', 'voltage', True))
FeedbackFloatVelocity = MessageEnumTraits(3, 'FeedbackFloatVelocity', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad/s', 'velocity', 'Velocity', 'velocity', True))
FeedbackFloatEffort = MessageEnumTraits(4, 'FeedbackFloatEffort', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('N*m', 'effort', 'Effort', 'effort', True))
FeedbackFloatVelocityCommand = MessageEnumTraits(5, 'FeedbackFloatVelocityCommand', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad/s', 'velocityCommand', 'VelocityCommand', 'velocity_command', True))
FeedbackFloatEffortCommand = MessageEnumTraits(6, 'FeedbackFloatEffortCommand', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('N*m', 'effortCommand', 'EffortCommand', 'effort_command', True))
FeedbackFloatDeflection = MessageEnumTraits(7, 'FeedbackFloatDeflection', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad', 'deflection', 'Deflection', 'deflection', True))
FeedbackFloatDeflectionVelocity = MessageEnumTraits(8, 'FeedbackFloatDeflectionVelocity', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad/s', 'deflectionVelocity', 'DeflectionVelocity', 'deflection_velocity', True))
FeedbackFloatMotorVelocity = MessageEnumTraits(9, 'FeedbackFloatMotorVelocity', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad/s', 'motorVelocity', 'MotorVelocity', 'motor_velocity', True))
FeedbackFloatMotorCurrent = MessageEnumTraits(10, 'FeedbackFloatMotorCurrent', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('A', 'motorCurrent', 'MotorCurrent', 'motor_current', True))
FeedbackFloatMotorSensorTemperature = MessageEnumTraits(11, 'FeedbackFloatMotorSensorTemperature', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('C', 'motorSensorTemperature', 'MotorSensorTemperature', 'motor_sensor_temperature', True))
FeedbackFloatMotorWindingCurrent = MessageEnumTraits(12, 'FeedbackFloatMotorWindingCurrent', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('A', 'motorWindingCurrent', 'MotorWindingCurrent', 'motor_winding_current', True))
FeedbackFloatMotorWindingTemperature = MessageEnumTraits(13, 'FeedbackFloatMotorWindingTemperature', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('C', 'motorWindingTemperature', 'MotorWindingTemperature', 'motor_winding_temperature', True))
FeedbackFloatMotorHousingTemperature = MessageEnumTraits(14, 'FeedbackFloatMotorHousingTemperature', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('C', 'motorHousingTemperature', 'MotorHousingTemperature', 'motor_housing_temperature', True))
FeedbackFloatBatteryLevel = MessageEnumTraits(15, 'FeedbackFloatBatteryLevel', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'batteryLevel', 'BatteryLevel', 'battery_level', True))
FeedbackFloatPwmCommand = MessageEnumTraits(16, 'FeedbackFloatPwmCommand', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'pwmCommand', 'PwmCommand', 'pwm_command', True))
FeedbackFloatField = EnumType('FeedbackFloatField', [ FeedbackFloatBoardTemperature, FeedbackFloatProcessorTemperature, FeedbackFloatVoltage, FeedbackFloatVelocity, FeedbackFloatEffort, FeedbackFloatVelocityCommand, FeedbackFloatEffortCommand, FeedbackFloatDeflection, FeedbackFloatDeflectionVelocity, FeedbackFloatMotorVelocity, FeedbackFloatMotorCurrent, FeedbackFloatMotorSensorTemperature, FeedbackFloatMotorWindingCurrent, FeedbackFloatMotorWindingTemperature, FeedbackFloatMotorHousingTemperature, FeedbackFloatBatteryLevel, FeedbackFloatPwmCommand, ])

# FeedbackHighResAngleField
FeedbackHighResAnglePosition = MessageEnumTraits(0, 'FeedbackHighResAnglePosition', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad', 'position', 'Position', 'position', True))
FeedbackHighResAnglePositionCommand = MessageEnumTraits(1, 'FeedbackHighResAnglePositionCommand', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad', 'positionCommand', 'PositionCommand', 'position_command', True))
FeedbackHighResAngleMotorPosition = MessageEnumTraits(2, 'FeedbackHighResAngleMotorPosition', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad', 'motorPosition', 'MotorPosition', 'motor_position', True))
FeedbackHighResAngleField = EnumType('FeedbackHighResAngleField', [ FeedbackHighResAnglePosition, FeedbackHighResAnglePositionCommand, FeedbackHighResAngleMotorPosition, ])

# FeedbackVector3fField
FeedbackVector3fAccelerometer = MessageEnumTraits(0, 'FeedbackVector3fAccelerometer', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
FeedbackVector3fGyro = MessageEnumTraits(1, 'FeedbackVector3fGyro', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
FeedbackVector3fArPosition = MessageEnumTraits(2, 'FeedbackVector3fArPosition', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
FeedbackVector3fField = EnumType('FeedbackVector3fField', [ FeedbackVector3fAccelerometer, FeedbackVector3fGyro, FeedbackVector3fArPosition, ])

# FeedbackQuaternionfField
FeedbackQuaternionfOrientation = MessageEnumTraits(0, 'FeedbackQuaternionfOrientation', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
FeedbackQuaternionfArOrientation = MessageEnumTraits(1, 'FeedbackQuaternionfArOrientation', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
FeedbackQuaternionfField = EnumType('FeedbackQuaternionfField', [ FeedbackQuaternionfOrientation, FeedbackQuaternionfArOrientation, ])

# FeedbackUInt64Field
FeedbackUInt64SequenceNumber = MessageEnumTraits(0, 'FeedbackUInt64SequenceNumber', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'sequenceNumber', 'SequenceNumber', 'sequence_number', True))
FeedbackUInt64ReceiveTime = MessageEnumTraits(1, 'FeedbackUInt64ReceiveTime', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('us', 'receiveTime', 'ReceiveTime', 'receive_time', True))
FeedbackUInt64TransmitTime = MessageEnumTraits(2, 'FeedbackUInt64TransmitTime', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('us', 'transmitTime', 'TransmitTime', 'transmit_time', True))
FeedbackUInt64HardwareReceiveTime = MessageEnumTraits(3, 'FeedbackUInt64HardwareReceiveTime', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('us', 'hardwareReceiveTime', 'HardwareReceiveTime', 'hardware_receive_time', True))
FeedbackUInt64HardwareTransmitTime = MessageEnumTraits(4, 'FeedbackUInt64HardwareTransmitTime', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('us', 'hardwareTransmitTime', 'HardwareTransmitTime', 'hardware_transmit_time', True))
FeedbackUInt64SenderId = MessageEnumTraits(5, 'FeedbackUInt64SenderId', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'senderId', 'SenderId', 'sender_id', True))
FeedbackUInt64Field = EnumType('FeedbackUInt64Field', [ FeedbackUInt64SequenceNumber, FeedbackUInt64ReceiveTime, FeedbackUInt64TransmitTime, FeedbackUInt64HardwareReceiveTime, FeedbackUInt64HardwareTransmitTime, FeedbackUInt64SenderId, ])

# FeedbackEnumField
FeedbackEnumTemperatureState = MessageEnumTraits(0, 'FeedbackEnumTemperatureState', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'temperatureState', 'TemperatureState', 'temperature_state', True))
FeedbackEnumMstopState = MessageEnumTraits(1, 'FeedbackEnumMstopState', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'mstopState', 'MstopState', 'mstop_state', True))
FeedbackEnumPositionLimitState = MessageEnumTraits(2, 'FeedbackEnumPositionLimitState', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionLimitState', 'PositionLimitState', 'position_limit_state', True))
FeedbackEnumVelocityLimitState = MessageEnumTraits(3, 'FeedbackEnumVelocityLimitState', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityLimitState', 'VelocityLimitState', 'velocity_limit_state', True))
FeedbackEnumEffortLimitState = MessageEnumTraits(4, 'FeedbackEnumEffortLimitState', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortLimitState', 'EffortLimitState', 'effort_limit_state', True))
FeedbackEnumCommandLifetimeState = MessageEnumTraits(5, 'FeedbackEnumCommandLifetimeState', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'commandLifetimeState', 'CommandLifetimeState', 'command_lifetime_state', True))
FeedbackEnumArQuality = MessageEnumTraits(6, 'FeedbackEnumArQuality', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'arQuality', 'ArQuality', 'ar_quality', True))
FeedbackEnumField = EnumType('FeedbackEnumField', [ FeedbackEnumTemperatureState, FeedbackEnumMstopState, FeedbackEnumPositionLimitState, FeedbackEnumVelocityLimitState, FeedbackEnumEffortLimitState, FeedbackEnumCommandLifetimeState, FeedbackEnumArQuality, ])


# FeedbackNumberedFloatField
FeedbackNumberedFloatDebug = MessageEnumTraits(0, 'FeedbackNumberedFloatDebug', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
FeedbackNumberedFloatField = EnumType('FeedbackNumberedFloatField', [ FeedbackNumberedFloatDebug, ])

# FeedbackIoBankField
FeedbackIoBankA = MessageEnumTraits(0, 'FeedbackIoBankA', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
FeedbackIoBankB = MessageEnumTraits(1, 'FeedbackIoBankB', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
FeedbackIoBankC = MessageEnumTraits(2, 'FeedbackIoBankC', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
FeedbackIoBankD = MessageEnumTraits(3, 'FeedbackIoBankD', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
FeedbackIoBankE = MessageEnumTraits(4, 'FeedbackIoBankE', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
FeedbackIoBankF = MessageEnumTraits(5, 'FeedbackIoBankF', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
FeedbackIoBankField = EnumType('FeedbackIoBankField', [ FeedbackIoBankA, FeedbackIoBankB, FeedbackIoBankC, FeedbackIoBankD, FeedbackIoBankE, FeedbackIoBankF, ])

# FeedbackLedField
FeedbackLedLed = MessageEnumTraits(0, 'FeedbackLedLed', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'led', 'Led', 'led', True))
FeedbackLedField = EnumType('FeedbackLedField', [ FeedbackLedLed, ])



# InfoFloatField
InfoFloatPositionKp = MessageEnumTraits(0, 'InfoFloatPositionKp', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionKp', 'PositionKp', 'position_kp', True))
InfoFloatPositionKi = MessageEnumTraits(1, 'InfoFloatPositionKi', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionKi', 'PositionKi', 'position_ki', True))
InfoFloatPositionKd = MessageEnumTraits(2, 'InfoFloatPositionKd', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionKd', 'PositionKd', 'position_kd', True))
InfoFloatPositionFeedForward = MessageEnumTraits(3, 'InfoFloatPositionFeedForward', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionFeedForward', 'PositionFeedForward', 'position_feed_forward', True))
InfoFloatPositionDeadZone = MessageEnumTraits(4, 'InfoFloatPositionDeadZone', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionDeadZone', 'PositionDeadZone', 'position_dead_zone', True))
InfoFloatPositionIClamp = MessageEnumTraits(5, 'InfoFloatPositionIClamp', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionIClamp', 'PositionIClamp', 'position_i_clamp', True))
InfoFloatPositionPunch = MessageEnumTraits(6, 'InfoFloatPositionPunch', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionPunch', 'PositionPunch', 'position_punch', True))
InfoFloatPositionMinTarget = MessageEnumTraits(7, 'InfoFloatPositionMinTarget', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionMinTarget', 'PositionMinTarget', 'position_min_target', True))
InfoFloatPositionMaxTarget = MessageEnumTraits(8, 'InfoFloatPositionMaxTarget', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionMaxTarget', 'PositionMaxTarget', 'position_max_target', True))
InfoFloatPositionTargetLowpass = MessageEnumTraits(9, 'InfoFloatPositionTargetLowpass', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionTargetLowpass', 'PositionTargetLowpass', 'position_target_lowpass', True))
InfoFloatPositionMinOutput = MessageEnumTraits(10, 'InfoFloatPositionMinOutput', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionMinOutput', 'PositionMinOutput', 'position_min_output', True))
InfoFloatPositionMaxOutput = MessageEnumTraits(11, 'InfoFloatPositionMaxOutput', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionMaxOutput', 'PositionMaxOutput', 'position_max_output', True))
InfoFloatPositionOutputLowpass = MessageEnumTraits(12, 'InfoFloatPositionOutputLowpass', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionOutputLowpass', 'PositionOutputLowpass', 'position_output_lowpass', True))
InfoFloatVelocityKp = MessageEnumTraits(13, 'InfoFloatVelocityKp', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityKp', 'VelocityKp', 'velocity_kp', True))
InfoFloatVelocityKi = MessageEnumTraits(14, 'InfoFloatVelocityKi', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityKi', 'VelocityKi', 'velocity_ki', True))
InfoFloatVelocityKd = MessageEnumTraits(15, 'InfoFloatVelocityKd', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityKd', 'VelocityKd', 'velocity_kd', True))
InfoFloatVelocityFeedForward = MessageEnumTraits(16, 'InfoFloatVelocityFeedForward', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityFeedForward', 'VelocityFeedForward', 'velocity_feed_forward', True))
InfoFloatVelocityDeadZone = MessageEnumTraits(17, 'InfoFloatVelocityDeadZone', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityDeadZone', 'VelocityDeadZone', 'velocity_dead_zone', True))
InfoFloatVelocityIClamp = MessageEnumTraits(18, 'InfoFloatVelocityIClamp', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityIClamp', 'VelocityIClamp', 'velocity_i_clamp', True))
InfoFloatVelocityPunch = MessageEnumTraits(19, 'InfoFloatVelocityPunch', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityPunch', 'VelocityPunch', 'velocity_punch', True))
InfoFloatVelocityMinTarget = MessageEnumTraits(20, 'InfoFloatVelocityMinTarget', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityMinTarget', 'VelocityMinTarget', 'velocity_min_target', True))
InfoFloatVelocityMaxTarget = MessageEnumTraits(21, 'InfoFloatVelocityMaxTarget', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityMaxTarget', 'VelocityMaxTarget', 'velocity_max_target', True))
InfoFloatVelocityTargetLowpass = MessageEnumTraits(22, 'InfoFloatVelocityTargetLowpass', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityTargetLowpass', 'VelocityTargetLowpass', 'velocity_target_lowpass', True))
InfoFloatVelocityMinOutput = MessageEnumTraits(23, 'InfoFloatVelocityMinOutput', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityMinOutput', 'VelocityMinOutput', 'velocity_min_output', True))
InfoFloatVelocityMaxOutput = MessageEnumTraits(24, 'InfoFloatVelocityMaxOutput', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityMaxOutput', 'VelocityMaxOutput', 'velocity_max_output', True))
InfoFloatVelocityOutputLowpass = MessageEnumTraits(25, 'InfoFloatVelocityOutputLowpass', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityOutputLowpass', 'VelocityOutputLowpass', 'velocity_output_lowpass', True))
InfoFloatEffortKp = MessageEnumTraits(26, 'InfoFloatEffortKp', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortKp', 'EffortKp', 'effort_kp', True))
InfoFloatEffortKi = MessageEnumTraits(27, 'InfoFloatEffortKi', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortKi', 'EffortKi', 'effort_ki', True))
InfoFloatEffortKd = MessageEnumTraits(28, 'InfoFloatEffortKd', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortKd', 'EffortKd', 'effort_kd', True))
InfoFloatEffortFeedForward = MessageEnumTraits(29, 'InfoFloatEffortFeedForward', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortFeedForward', 'EffortFeedForward', 'effort_feed_forward', True))
InfoFloatEffortDeadZone = MessageEnumTraits(30, 'InfoFloatEffortDeadZone', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortDeadZone', 'EffortDeadZone', 'effort_dead_zone', True))
InfoFloatEffortIClamp = MessageEnumTraits(31, 'InfoFloatEffortIClamp', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortIClamp', 'EffortIClamp', 'effort_i_clamp', True))
InfoFloatEffortPunch = MessageEnumTraits(32, 'InfoFloatEffortPunch', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortPunch', 'EffortPunch', 'effort_punch', True))
InfoFloatEffortMinTarget = MessageEnumTraits(33, 'InfoFloatEffortMinTarget', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortMinTarget', 'EffortMinTarget', 'effort_min_target', True))
InfoFloatEffortMaxTarget = MessageEnumTraits(34, 'InfoFloatEffortMaxTarget', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortMaxTarget', 'EffortMaxTarget', 'effort_max_target', True))
InfoFloatEffortTargetLowpass = MessageEnumTraits(35, 'InfoFloatEffortTargetLowpass', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortTargetLowpass', 'EffortTargetLowpass', 'effort_target_lowpass', True))
InfoFloatEffortMinOutput = MessageEnumTraits(36, 'InfoFloatEffortMinOutput', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortMinOutput', 'EffortMinOutput', 'effort_min_output', True))
InfoFloatEffortMaxOutput = MessageEnumTraits(37, 'InfoFloatEffortMaxOutput', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortMaxOutput', 'EffortMaxOutput', 'effort_max_output', True))
InfoFloatEffortOutputLowpass = MessageEnumTraits(38, 'InfoFloatEffortOutputLowpass', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortOutputLowpass', 'EffortOutputLowpass', 'effort_output_lowpass', True))
InfoFloatSpringConstant = MessageEnumTraits(39, 'InfoFloatSpringConstant', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('N/m', 'springConstant', 'SpringConstant', 'spring_constant', True))
InfoFloatVelocityLimitMin = MessageEnumTraits(40, 'InfoFloatVelocityLimitMin', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad/s', 'velocityLimitMin', 'VelocityLimitMin', 'velocity_limit_min', True))
InfoFloatVelocityLimitMax = MessageEnumTraits(41, 'InfoFloatVelocityLimitMax', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad/s', 'velocityLimitMax', 'VelocityLimitMax', 'velocity_limit_max', True))
InfoFloatEffortLimitMin = MessageEnumTraits(42, 'InfoFloatEffortLimitMin', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('N*m', 'effortLimitMin', 'EffortLimitMin', 'effort_limit_min', True))
InfoFloatEffortLimitMax = MessageEnumTraits(43, 'InfoFloatEffortLimitMax', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('N*m', 'effortLimitMax', 'EffortLimitMax', 'effort_limit_max', True))
InfoFloatField = EnumType('InfoFloatField', [ InfoFloatPositionKp, InfoFloatPositionKi, InfoFloatPositionKd, InfoFloatPositionFeedForward, InfoFloatPositionDeadZone, InfoFloatPositionIClamp, InfoFloatPositionPunch, InfoFloatPositionMinTarget, InfoFloatPositionMaxTarget, InfoFloatPositionTargetLowpass, InfoFloatPositionMinOutput, InfoFloatPositionMaxOutput, InfoFloatPositionOutputLowpass, InfoFloatVelocityKp, InfoFloatVelocityKi, InfoFloatVelocityKd, InfoFloatVelocityFeedForward, InfoFloatVelocityDeadZone, InfoFloatVelocityIClamp, InfoFloatVelocityPunch, InfoFloatVelocityMinTarget, InfoFloatVelocityMaxTarget, InfoFloatVelocityTargetLowpass, InfoFloatVelocityMinOutput, InfoFloatVelocityMaxOutput, InfoFloatVelocityOutputLowpass, InfoFloatEffortKp, InfoFloatEffortKi, InfoFloatEffortKd, InfoFloatEffortFeedForward, InfoFloatEffortDeadZone, InfoFloatEffortIClamp, InfoFloatEffortPunch, InfoFloatEffortMinTarget, InfoFloatEffortMaxTarget, InfoFloatEffortTargetLowpass, InfoFloatEffortMinOutput, InfoFloatEffortMaxOutput, InfoFloatEffortOutputLowpass, InfoFloatSpringConstant, InfoFloatVelocityLimitMin, InfoFloatVelocityLimitMax, InfoFloatEffortLimitMin, InfoFloatEffortLimitMax, ])

# InfoHighResAngleField
InfoHighResAnglePositionLimitMin = MessageEnumTraits(0, 'InfoHighResAnglePositionLimitMin', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad', 'positionLimitMin', 'PositionLimitMin', 'position_limit_min', True))
InfoHighResAnglePositionLimitMax = MessageEnumTraits(1, 'InfoHighResAnglePositionLimitMax', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('rad', 'positionLimitMax', 'PositionLimitMax', 'position_limit_max', True))
InfoHighResAngleField = EnumType('InfoHighResAngleField', [ InfoHighResAnglePositionLimitMin, InfoHighResAnglePositionLimitMax, ])




# InfoEnumField
InfoEnumControlStrategy = MessageEnumTraits(0, 'InfoEnumControlStrategy', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'controlStrategy', 'ControlStrategy', 'control_strategy', True))
InfoEnumCalibrationState = MessageEnumTraits(1, 'InfoEnumCalibrationState', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'calibrationState', 'CalibrationState', 'calibration_state', True))
InfoEnumMstopStrategy = MessageEnumTraits(2, 'InfoEnumMstopStrategy', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'mstopStrategy', 'MstopStrategy', 'mstop_strategy', True))
InfoEnumMinPositionLimitStrategy = MessageEnumTraits(3, 'InfoEnumMinPositionLimitStrategy', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'minPositionLimitStrategy', 'MinPositionLimitStrategy', 'min_position_limit_strategy', True))
InfoEnumMaxPositionLimitStrategy = MessageEnumTraits(4, 'InfoEnumMaxPositionLimitStrategy', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'maxPositionLimitStrategy', 'MaxPositionLimitStrategy', 'max_position_limit_strategy', True))
InfoEnumField = EnumType('InfoEnumField', [ InfoEnumControlStrategy, InfoEnumCalibrationState, InfoEnumMstopStrategy, InfoEnumMinPositionLimitStrategy, InfoEnumMaxPositionLimitStrategy, ])

# InfoBoolField
InfoBoolPositionDOnError = MessageEnumTraits(0, 'InfoBoolPositionDOnError', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'positionDOnError', 'PositionDOnError', 'position_d_on_error', True))
InfoBoolVelocityDOnError = MessageEnumTraits(1, 'InfoBoolVelocityDOnError', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'velocityDOnError', 'VelocityDOnError', 'velocity_d_on_error', True))
InfoBoolEffortDOnError = MessageEnumTraits(2, 'InfoBoolEffortDOnError', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'effortDOnError', 'EffortDOnError', 'effort_d_on_error', True))
InfoBoolAccelIncludesGravity = MessageEnumTraits(3, 'InfoBoolAccelIncludesGravity', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'accelIncludesGravity', 'AccelIncludesGravity', 'accel_includes_gravity', True))
InfoBoolField = EnumType('InfoBoolField', [ InfoBoolPositionDOnError, InfoBoolVelocityDOnError, InfoBoolEffortDOnError, InfoBoolAccelIncludesGravity, ])



# InfoLedField
InfoLedLed = MessageEnumTraits(0, 'InfoLedLed', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'led', 'Led', 'led', True))
InfoLedField = EnumType('InfoLedField', [ InfoLedLed, ])

# InfoStringField
InfoStringName = MessageEnumTraits(0, 'InfoStringName', allow_broadcast=False, not_bcastable_reason='Cannot set same name for all modules in a group.', yields_dict=None, field_details=None)
InfoStringFamily = MessageEnumTraits(1, 'InfoStringFamily', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
InfoStringSerial = MessageEnumTraits(2, 'InfoStringSerial', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=None)
InfoStringField = EnumType('InfoStringField', [ InfoStringName, InfoStringFamily, InfoStringSerial, ])

# InfoFlagField
InfoFlagSaveCurrentSettings = MessageEnumTraits(0, 'InfoFlagSaveCurrentSettings', allow_broadcast=True, not_bcastable_reason=None, yields_dict=None, field_details=FieldDetails('None', 'saveCurrentSettings', 'SaveCurrentSettings', 'save_current_settings', True))
InfoFlagField = EnumType('InfoFlagField', [ InfoFlagSaveCurrentSettings, ])



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  HEBI Core python API - Copyright 2018 HEBI Robotics
#  See https://hebi.us/softwarelicense for license details
#
# -----------------------------------------------------------------------------


class HEBI_Exception(Exception):
  """
  Internal exception with HEBI error message
  """

  def __init__(self, errcode, message=''):
    self._errcode = errcode
    self._message = message

  def __str__(self):
    return '{0} ({1})'.format(self._message, self._errors.get(int(self._errcode),
                                                    'unknown error'))

  def __repr__(self):
    return str(self)

  _errors = {
    0x00 : 'no error',
    0x01 : 'invalid argument passed to API call',
    0x02 : 'provided buffer passed to API call is too small',
    0x03 : 'field value not set',
    0x04 : 'API call failed',
    0x05 : 'provided argument passed to API call is out of expected range',
  }



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  HEBI Core python API - Copyright 2018 HEBI Robotics
#  See https://hebi.us/softwarelicense for license details
#
# -----------------------------------------------------------------------------


import os
import platform
from sys import intern as intern_string
import weakref


################################################################################
# Classes
################################################################################


class Counter(object):
  """
  Counter class. For internal use - do not use directly.
  """

  __slots__ = ['_counter']

  def __init__(self):
    self._counter = 1

  def decrement(self):
    self._counter = self._counter - 1
    return self._counter

  def increment(self):
    self._counter = self._counter + 1
    return self._counter

  def clear(self):
    self._counter = 0

  def __gt__(self, val):
    return self._counter > val

  def __lt__(self, val):
    return self._counter < val

  @property
  def count(self):
    return self._counter


class CaseInvariantString(object):
  """
  Represents an immutable string with a custom hash implementation and case invariant comparison
  """

  __slots__ = ['__hash', '__lower_val', '__val']

  def __init__(self, val):
    val = str(val)
    self.__val = val
    self.__lower_val = val.strip().lower()
    self.__hash = hash(self.__lower_val)

  @property
  def value(self):
    return self.__lower_val

  def __hash__(self):
    return self.__hash

  def __eq__(self, other):
    if type(other) is CaseInvariantString:
      return self.__lower_val == other.value
    return str(other).lower() != self.__lower_val

  def __ne__(self, other):
    if type(other) is CaseInvariantString:
      return self.__lower_val != other.value
    return str(other).lower() != self.__lower_val

  def __str__(self):
    return self.__val

  def __repr__(self):
    return self.__val


################################################################################
# Pretty Strings
################################################################################


def truncate_with_r_justify(data, length):
  data = str(data)
  data_len = len(data)

  if data_len > length:
    # Truncate the string to [length-3], then add ellipsis
    data = data[0:length-3] + '...'

  fmt_str = '{' + ':<{0}'.format(length) + '}'
  return fmt_str.format(data[0:length])


def lookup_table_string(lookup_entries):
  length = len(lookup_entries)
  if length < 1:
    return 'No modules on network'
  max_module_length = 6
  max_family_length = 16
  max_name_length   = 14
  ret = 'Module  Family            Name          \n'
  ret = ret +\
        '------  ----------------  --------------\n'
  for i, entry in enumerate(lookup_entries):
    module_str = truncate_with_r_justify(i, max_module_length)
    family_str = truncate_with_r_justify(entry.family, max_family_length)
    name_str   = truncate_with_r_justify(entry.name, max_name_length)
    ret = ret +\
      '{0}  {1}  {2}\n'.format(module_str, family_str, name_str)
  return ret


################################################################################
# File system methods
################################################################################


def safe_mkdirs(name):
  """
  Wraps `os.makedirs`, while additionally handling edge cases
  """
  return os.makedirs(os.path.realpath(name), exist_ok=True)


################################################################################
# Compatibility Layer
################################################################################


__is_pypy = platform.python_implementation().lower() == 'pypy'


def is_pypy():
  return __is_pypy




import sys
if sys.version_info[0] == 3 and sys.version_info[1] >= 3:
  from time import perf_counter as _get_time
else:
  from time import time as _get_time
from time import strftime


import numpy as np
import threading


class _HebiProfilerTLS_Data(object):

  __slots__ = ['_func_name', '_start_time', '_func_stats']

  def __init__(self):
    self._func_name = None
    self._start_time = None
    self._func_stats = dict()
    from .loader import _c_api_functions
    for entry in _c_api_functions:
      self._func_stats[entry.name] = list()

  @property
  def func_name(self):
    return self._func_name

  @func_name.setter
  def func_name(self, value):
    self._func_name = value

  @property
  def start_time(self):
    return self._start_time

  @start_time.setter
  def start_time(self, value):
    self._start_time = value

  @property
  def func_stats(self):
    return self._func_stats


class _HebiProfilerTLS(threading.local):
  """
  Thread local storage for the internal profiler.
  """
  def __init__(self):
    self._data = _HebiProfilerTLS_Data()

  @property
  def data(self):
    return self._data


_profiler_tls = _HebiProfilerTLS()


class _SimpleHEBIProfiler(object):
  """
  A simple profiler used to profile how long calls into HEBI C API take.
  Note: this works on a per thread basis and is currently an internal API.
  """
  def __init__(self):
    pass

  def enter(self, func_name):
    tls_data = _profiler_tls.data
    tls_data.func_name = func_name
    tls_data.start_time = _get_time()

  def exit(self):
    tls_data = _profiler_tls.data
    tls_data.func_stats[tls_data.func_name].append(_get_time()-tls_data.start_time)

  def _write_entry(self, f, func_name):
    tls_data = _profiler_tls.data
    entries = tls_data.func_stats[func_name]
    if len(entries) == 0:
      return
    entries = np.array(entries)
    min_val = np.amin(entries)
    max_val = np.amax(entries)
    mean = np.mean(entries)
    stddev = np.std(entries)
    count = entries.size
    pct_90 = np.percentile(entries, 90)
    pct_95 = np.percentile(entries, 95)
    pct_99 = np.percentile(entries, 99)
    cumulative = np.sum(entries)
    f.write('{0},{1},{2},{3},{4},{5},{6},{7},{8},{9}\n'.format(func_name, count,
                                                               cumulative, mean,
                                                               min_val, max_val,
                                                               stddev, pct_90,
                                                               pct_95, pct_99))

  def dump_stats(self, path=None):
    if path is None:
      thd = threading.current_thread()
      path = '{0}_{1}-{2}_c_profile.csv'.format(strftime('%m_%d_%y_%H-%M-%S'), thd.name, thd.ident)

    tls_data = _profiler_tls.data
    with open(path, 'w') as f:
      f.write('function,count,cumulative,avg,min,max,stddev,90pct,95pct,99pct\n')
      for func_name in tls_data.func_stats.keys():
        self._write_entry(f, func_name)


_C_API_profiler = _SimpleHEBIProfiler()


def enter_c_func(func_name):
  _C_API_profiler.enter(func_name)


def exit_c_func():
  _C_API_profiler.exit()


def dump_c_api_stats(path=None):
  _C_API_profiler.dump_stats(path)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  HEBI Core python API - Copyright 2018 HEBI Robotics
#  See https://hebi.us/softwarelicense for license details
#
# -----------------------------------------------------------------------------


from ctypes import CFUNCTYPE, c_size_t, c_void_p, c_char_p, byref
import os

from .log_file import LogFile
from .ffi._message_types import GroupFeedback, GroupInfo
from .ffi import ctypes_func_defs as api
from .ffi import enums
from .ffi.wrappers import UnmanagedObject, UnmanagedSharedObject
from .utils import safe_mkdirs
from . import type_utils

_FeedbackHandlerFunction = CFUNCTYPE(None, c_void_p, c_void_p)


class Group(object):
  """
  Represents a group of physical HEBI modules, and allows Command,
  Feedback, and Info objects to be sent to and recieved from the hardware.
  """

  __slots__ = ['__delegate']

  DEFAULT_TIMEOUT_MS = 500
  """
  The default timeout (in milliseconds)
  """

  def __init__(self, delegate):
    """
    This is created internally. Do not instantiate directly.
    """
    self.__delegate = delegate

  def __repr__(self):
    if self.__delegate.finalized:
      return 'Finalized group (size: {})'.format(self.size)
    feedback_freq = self.feedback_frequency
    # command_lifetime is in milliseconds - convert to seconds
    cmd_lifetime = float(self.command_lifetime) * 0.001
    num_modules = self.size
    
    justify_size = 20
    justify_fmt  = '{' + ':>{0}'.format(justify_size) + '}'

    fbk_str = '{} [Hz]'.format(feedback_freq)
    cmd_str = '{} [s]'.format(cmd_lifetime)
    sze_str = '{}'.format(num_modules)

    ret = justify_fmt.format('feedback_frequency: ') + fbk_str + '\n'
    ret = ret +\
      justify_fmt.format('command_lifetime: ') + cmd_str + '\n'
    ret = ret +\
      justify_fmt.format('size: ') + sze_str

    # XXX: We should really add the module info table, to match MATLAB
    return ret

  def __str__(self):
    feedback_freq = self.feedback_frequency
    # command_lifetime is in milliseconds - convert to seconds
    cmd_lifetime = float(self.command_lifetime) * 0.001
    num_modules = self.size
    return 'Group(feedback_frequency={}, '.format(feedback_freq) +\
      'command_lifetime={}, '.format(cmd_lifetime) +\
      'size={})'.format(num_modules)

  @property
  def size(self):
    """
    The number of modules in the group.

    :return: size of the group
    :rtype:  int
    """
    return self.__delegate.size

  @property
  def feedback_frequency(self):
    """
    The frequency of the internal feedback request + callback thread.

    :return: the frequency
    :rtype:  float
    """
    return self.__delegate.feedback_frequency

  @feedback_frequency.setter
  def feedback_frequency(self, value):
    """
    Sets the frequency of the internal feedback request + callback thread.

    :param value: the new frequency, in hertz
    :type value:  float
    """
    self.__delegate.feedback_frequency = value

  @property
  def command_lifetime(self):
    """
    The command lifetime for the modules in this group.

    :return: the command lifetime
    :rtype:  int
    """
    return self.__delegate.command_lifetime

  @command_lifetime.setter
  def command_lifetime(self, value):
    """
    Set the command lifetime for the modules in this group.

    This parameter defines how long a module will execute a command set
    point sent to it. Note the commands from other systems/applications
    are ignored during this time. A value of ``0`` indicates commands last
    forever, and there is no lockout behavior.

    :param value: The command lifetime, in milliseconds
    :type value:  int
    """
    self.__delegate.command_lifetime = value

  def send_command(self, group_command):
    """
    Send a command to the given group without requesting an acknowledgement.
    This is appropriate for high-frequency applications.

    :param group_command: The command to send
    :type group_command:  GroupCommand

    :return: ``True`` if succeeded, ``False`` otherwise
    :rtype:  bool
    """
    return self.__delegate.send_command(group_command)

  def send_command_with_acknowledgement(self, group_command, timeout_ms=None):
    """
    Send a command to the given group, requesting an acknowledgement
    of transmission to be sent back.

    Note: A non-true return does not indicate a specific failure, and may
    result from an error while sending or simply a timeout/dropped response
    packet after a successful transmission.

    :param group_command: The command to send
    :type group_command:  GroupCommand

    :param timeout_ms: The maximum amount of time to wait, in milliseconds.
                       This is an optional parameter with a default value of
                       :attr:`Group.DEFAULT_TIMEOUT_MS`.
    :type timeout_ms:  int

    :return: ``True`` if succeeded, ``False`` otherwise
    :rtype:  bool
    """
    return self.__delegate.send_command_with_acknowledgement(group_command,
                                                             timeout_ms)

  def send_feedback_request(self):
    """
    Requests feedback from the group.

    Sends a background request to the modules in the group; if/when all modules
    return feedback, any associated handler functions are called. This returned
    feedback is also stored to be returned by the next call to
    :meth:`.get_next_feedback` (any previously returned data is discarded).
    
    :return: ``True`` if succeeded, ``False`` otherwise
    :rtype:  bool
    """
    return self.__delegate.send_feedback_request()

  def get_next_feedback(self, timeout_ms=None, reuse_fbk=None):
    """
    Returns the most recently stored feedback from a sent feedback request, or
    the next one received (up to the requested timeout).

    Note that a feedback request can be sent either with the
    send_feedback_request function, or by setting a background feedback
    frequency with :attr:`.feedback_frequency`.

    :param timeout_ms: The maximum amount of time to wait, in milliseconds.
                       This is an optional parameter with a default value of
                       :attr:`Group.DEFAULT_TIMEOUT_MS`.
    :type timeout_ms:  int

    :param reuse_fbk: An optional parameter which can be used to reuse
                      an existing GroupFeedback instance. It is recommended
                      to provide this parameter inside a repetitive loop,
                      as reusing feedback instances results in substantially
                      fewer heap allocations.
    :type reuse_fbk:  GroupFeedback

    :return: The most recent feedback, provided one became available before the
             timeout. ``None`` is returned if there was no available feedback.
    :rtype:  GroupFeedback
    """
    return self.__delegate.get_next_feedback(timeout_ms, reuse_fbk)

  def request_info(self, timeout_ms=None, reuse_info=None):
    """
    Request info from the group, and store it in the passed-in info object.

    :param timeout_ms: The maximum amount of time to wait, in milliseconds.
                       This is an optional parameter with a default value of
                       :attr:`Group.DEFAULT_TIMEOUT_MS`.
    :type timeout_ms:  int

    :param reuse_info: An optional parameter which can be used to reuse
                       an existing GroupInfo instance. It is recommended
                       to provide this parameter inside a repetitive loop,
                       as reusing info instances results in substantially
                       fewer heap allocations.
    :type reuse_info:  GroupInfo

    :return: the updated info on success, ``None`` otherwise
    :rtype:  GroupInfo
    """
    return self.__delegate.request_info(timeout_ms, reuse_info)

  def start_log(self, directory=None, name=None, mkdirs=False):
    """
    Start logging information and feedback from this group.

    If a log file was already started before this (and not stopped using :meth:`.stop_log`),
    then that file will be gracefully closed.

    Note that the `directory` folder must exist if `mkdirs` is not ``True``

    :param directory: Optional directory into which the log file will be created.
                      If ``None``, the process' current working directory is used.
    :type directory:  str

    :param name: Optional name of the log file.
                 If ``None``, a name will be generated using the time
                 at which this function was invoked.
    :type name:  str

    :param mkdirs: Optional flag denoting if the base path represented by `directory` should be created,
                   if they do not exist
    :type mkdirs:  bool

    :return: The path, including the file, of the log file. Never ``None``.
    :rtype:  str

    :raises IOError: If the group could not start logging
    """
    if mkdirs and directory is not None:
      safe_mkdirs(directory)
    ret = self.__delegate.start_log(directory, name)
    if ret is None:
      if directory is not None and not mkdirs and not os.path.isdir(directory):
        # Hueristic: 
        # Try to make it clear to the user that the directory in which they desired
        # to log does not exist, when they did not set the `mkdirs` flag
        raise IOError("Could not start logging because directory '{}' does not exist and 'mkdirs=False'".format(directory))

      raise IOError("Could not start logging")
    return ret

  def stop_log(self):
    """
    Stop logging data into the last opened log file.
    If no log file was opened, None will be returned.
    If an error occurs while stopping the previously started
    log file, None will be returned.

    :return: a LogFile object on success, or None
    :rtype:  LogFile
    """
    return self.__delegate.stop_log()

  def add_feedback_handler(self, handler):
    """
    Adds a handler function to be called by the internal feedback request thread.

    Note that this function must execute very quickly:
    If a handler takes more time than the reciprocal of the feedback
    thread frequency, the thread will saturate in feedback events to dispatch.
    This may cause feedback packets to be dropped from handler dispatch,
    or delayed invocation of the feedback handlers.

    :param handler: A function which is able to accept a single argument
    """
    self.__delegate.add_feedback_handler(handler)

  def clear_feedback_handlers(self):
    """
    Removes all feedback handlers presently added.
    """
    self.__delegate.clear_feedback_handlers()


class GroupDelegate(UnmanagedObject):
  """
  Delegate for Group
  """

  __slots__ = ['_c_callback', '_feedback_callbacks', '_number_of_modules', '__weakref__']

  __instances = list()

  @staticmethod
  def destroy_all_instances():
    for entry in GroupDelegate.__instances:
      try:
        entry().force_delete()
      except:
        pass

  def __parse_to(self, timeout_ms):
    if timeout_ms is None:
      return Group.DEFAULT_TIMEOUT_MS
    else:
      try:
        return int(timeout_ms)
      except:
        raise ValueError('timeout_ms must be a number')

  def __feedback_handler(self, c_fbk, c_data):

    class FlyweightGroupFeedback(GroupFeedback):
      __slots__ = []
      def __init__(self, number_of_modules, internal):
        # Explicitly skip over GroupFeedback constructor
        UnmanagedSharedObject.__init__(self, internal=internal)
        self._initialize(number_of_modules)


    feedback = FlyweightGroupFeedback(self._number_of_modules, c_fbk)
    for entry in self._feedback_callbacks:
      entry(feedback)
    # Don't allow any dangling references
    feedback._holder.force_delete()

  def __setup_feedback_handler(self):
    fbk_handler = self.__feedback_handler
    c_callback = _FeedbackHandlerFunction(fbk_handler)
    self._c_callback = c_callback
    api.hebiGroupRegisterFeedbackHandler(self, c_callback, c_void_p(0))

  def __init__(self, internal):

    def deleter(internal):
      api.hebiGroupSetFeedbackFrequencyHz(internal, 0.0)
      api.hebiGroupClearFeedbackHandlers(internal)
      # If the group is logging, stop.
      c_log = api.hebiGroupStopLog(internal)
      if c_log is not None:
        # Release the log file, if it was created.
        api.hebiLogFileRelease(c_log)
      api.hebiGroupRelease(internal)

    super(GroupDelegate, self).__init__(internal, on_delete=deleter)

    from weakref import ref
    GroupDelegate.__instances.append(ref(self))

    self._number_of_modules = int(api.hebiGroupGetSize(internal))
    self._feedback_callbacks = list()

  @property
  def size(self):
    return self._number_of_modules

  @property
  def feedback_frequency(self):
    return float(api.hebiGroupGetFeedbackFrequencyHz(self))

  @feedback_frequency.setter
  def feedback_frequency(self, value):
    try:
      value = float(value)
    except:
      raise ValueError("frequency must be a number")
    if api.hebiGroupSetFeedbackFrequencyHz(self, value) != enums.StatusSuccess:
      raise RuntimeError('Could not set feedback frequency to {}'.format(value))

  @property
  def command_lifetime(self):
    return int(api.hebiGroupGetCommandLifetime(self))

  @command_lifetime.setter
  def command_lifetime(self, value):
    try:
      value = int(value)
    except:
      raise ValueError("lifetime must be a number")
    if api.hebiGroupSetCommandLifetime(self, value) != enums.StatusSuccess:
      raise RuntimeError('Could not set command lifetime to {}'.format(value))

  def send_command(self, group_command):
    return api.hebiGroupSendCommand(self, group_command) == enums.StatusSuccess

  def send_command_with_acknowledgement(self, group_command, timeout_ms=None):
    timeout_ms = self.__parse_to(timeout_ms)
    return (
      api.hebiGroupSendCommandWithAcknowledgement(self, group_command, timeout_ms)
      == enums.StatusSuccess)

  def send_feedback_request(self):
    return api.hebiGroupSendFeedbackRequest(self) == enums.StatusSuccess

  def get_next_feedback(self, timeout_ms=None, reuse_fbk=None):
    timeout_ms = self.__parse_to(timeout_ms)
    if reuse_fbk:
      feedback = GroupFeedback(self._number_of_modules, reuse_fbk)
    else:
      feedback = GroupFeedback(self._number_of_modules)

    res = api.hebiGroupGetNextFeedback(self, feedback, timeout_ms)
    if res == enums.StatusSuccess:
      return feedback
    else:
      return None

  def request_info(self, timeout_ms=None, reuse_info=None):
    timeout_ms = self.__parse_to(timeout_ms)
    if reuse_info is not None:
      info = GroupInfo(self._number_of_modules)
    else:
      info = GroupInfo(self._number_of_modules, reuse_info)

    res = api.hebiGroupRequestInfo(self, info, timeout_ms)
    if res == enums.StatusSuccess:
      return info
    else:
      return None

  def start_log(self, directory=None, name=None):
    if directory is not None:
      directory = str(directory)
      c_directory = type_utils.create_string_buffer_compat(directory, len(directory))
    else:
      c_directory = c_char_p(None)

    if name is not None:
      name = str(name)
      if '.' not in name:
        name = name + '.hebilog'
      c_name = type_utils.create_string_buffer_compat(name, len(name))
    else:
      c_name = c_char_p(None)

    c_string_ret = c_void_p(0)
    res = api.hebiGroupStartLog(self, c_directory, c_name, byref(c_string_ret))

    if res == enums.StatusSuccess:
      c_str_length = c_size_t(0)
      api.hebiStringGetString(c_string_ret, c_char_p(None), byref(c_str_length))

      if c_str_length.value == 0: # Unknown error
        api.hebiStringRelease(c_string_ret)
        raise RuntimeError(
          'Underlying C API call hebiStringGetString'
          'returned length of zero')

      c_str_buffer = type_utils.create_string_buffer_compat(c_str_length.value)
      api.hebiStringGetString(c_string_ret, c_str_buffer, byref(c_str_length))

      ret = c_str_buffer.value.decode('utf-8')
      api.hebiStringRelease(c_string_ret)
      return ret

    return None

  def stop_log(self):
    c_log = api.hebiGroupStopLog(self)
    if c_log is None:
      return None
    return LogFile(c_log)

  def add_feedback_handler(self, handler):
    if not callable(handler):
      raise ValueError('handler was not callable')
    if handler in self._feedback_callbacks:
      return
    self._feedback_callbacks.append(handler)
    if len(self._feedback_callbacks) == 1:
      self.__setup_feedback_handler()


  def clear_feedback_handlers(self):
    self._feedback_callbacks = list()
    api.hebiGroupClearFeedbackHandlers(self)


def create_imitation_group(size):
  if type(size) is not int:
    raise TypeError('size must be an int')
  if size < 1:
    raise ValueError('size must be greater than zero')

  ret = Group(GroupDelegate(api.hebiGroupCreateImitation(size)))
  return ret


# Destroy all GroupDelegate objects on exit.
# This ensures that the feedback handlers cannot be invoked while Python is finalizing.
# This allows Python to exit gracefully, while also additionally allowing any potentially
# open log files to gracefully be closed and validated (preventing corruption).
from atexit import register as __register
__register(GroupDelegate.destroy_all_instances)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  HEBI Core python API - Copyright 2018 HEBI Robotics
#  See https://hebi.us/softwarelicense for license details
#
# -----------------------------------------------------------------------------


import math
import numpy as np
import ctypes
from ctypes import (byref, create_string_buffer)

from ..graphics import Color, color_from_int, string_to_color
from .enums import StatusSuccess
from . import ctypes_func_defs as api
from .wrappers import WeakReferenceContainer # TODO: fix import
from .. import type_utils
from ..type_utils import decode_string_buffer as decode_str # TODO: fix import
from ..type_utils import create_string_buffer_compat as create_str # TODO: fix import

from hebi._internal.ffi.ctypes_defs import HebiVector3f, HebiQuaternionf, HebiHighResAngleStruct
from hebi._internal.ffi.ctypes_utils import to_float_ptr, to_double_ptr, cast_to_float_ptr

from numpy.ctypeslib import as_array as _ctypes_to_ndarray


################################################################################
# Pretty Printers
################################################################################


def _fmt_float_array(array):
  return '[' + ', '.join(['{:.2f}'.format(i) for i in array]) + ']'


def _numbered_float_repr(c, enum_type):
  try:
    enum_name = enum_type.name
  except:
    enum_name = enum_type
  desc = 'Numbered float (Enumeration {}):\n'.format(enum_name)
  try:
    container = c._get_ref()
    return desc +\
      '  float1: {}\n'.format(_fmt_float_array(c.float1)) +\
      '  float2: {}\n'.format(_fmt_float_array(c.float2)) +\
      '  float3: {}\n'.format(_fmt_float_array(c.float3)) +\
      '  float4: {}\n'.format(_fmt_float_array(c.float4)) +\
      '  float5: {}\n'.format(_fmt_float_array(c.float5)) +\
      '  float6: {}\n'.format(_fmt_float_array(c.float6)) +\
      '  float7: {}\n'.format(_fmt_float_array(c.float7)) +\
      '  float8: {}\n'.format(_fmt_float_array(c.float8)) +\
      '  float9: {}\n'.format(_fmt_float_array(c.float9))
  except:
    return desc + '  <Group message was finalized>'


def _fmt_io_bank_pin(pins, indent=4):
  indent_str = ''.join([' '] * indent)
  pins_i = pins[0]
  pins_f = pins[1]

  pins_i_str = '[' + ', '.join(['{0:9g}'.format(entry) for entry in pins_i]) + ']'
  pins_f_str = '[' + ', '.join(['{0:9.8g}'.format(entry) for entry in pins_f]) + ']'

  return '{}Int:   {}\n'.format(indent_str, pins_i_str) +\
         '{}Float: {}'.format(indent_str, pins_f_str)


def _io_bank_repr(bank_container, bank, bank_readable):
  try:
    enum_name = bank.name
  except:
    enum_name = bank
  desc = 'IO Bank \'{}\' (Enumeration {}):\n'.format(bank_readable, enum_name)
  try:
    io_container = bank_container._get_ref()
  except:
    # Handles the case where IO Container object was finalized already 
    return desc + "  <IO Container was finalized>"

  # TODO: Maybe replace non-existent pins with `N/A` instead of `nan` or `0`?
  def get_fmt_pin(pin):
    return _fmt_io_bank_pin((io_container.get_int(bank, pin), io_container.get_float(bank, pin)))

  return desc +\
    '  Pin 1:\n{}\n'.format(get_fmt_pin(1)) +\
    '  Pin 2:\n{}\n'.format(get_fmt_pin(2)) +\
    '  Pin 3:\n{}\n'.format(get_fmt_pin(3)) +\
    '  Pin 4:\n{}\n'.format(get_fmt_pin(4)) +\
    '  Pin 5:\n{}\n'.format(get_fmt_pin(5)) +\
    '  Pin 6:\n{}\n'.format(get_fmt_pin(6)) +\
    '  Pin 7:\n{}\n'.format(get_fmt_pin(7)) +\
    '  Pin 8:\n{}\n'.format(get_fmt_pin(8))


def _io_repr(io):
  try:
    container = io._get_ref()
    return 'IO Banks: [A, B, C, D, E, F]\n' +\
      '{}\n'.format(io.a) +\
      '{}\n'.format(io.b) +\
      '{}\n'.format(io.c) +\
      '{}\n'.format(io.d) +\
      '{}\n'.format(io.e) +\
      str(io.f)
  except:
    return 'IO Banks: [A, B, C, D, E, F]\n  <Group message was finalized>'


def _led_repr(led, led_field):
  try:
    enum_name = led_field.name
  except:
    enum_name = led_field
  desc = 'LED (Enumeration {}):\n'.format(enum_name)
  try:
    container = led._get_ref()
    colors = led.color
    return desc + '  [' + ', '.join([repr(color) for color in colors]) + ']'
  except:
    return desc + '  <Group message was finalized>'


################################################################################
# Numbered Fields
################################################################################

################################################################################
# `has` creators
################################################################################


def __create_has_func(refs, bit_index, has):
  size = len(refs)

  def ret_has():
    _tls.ensure_capacity(size)
    output = _tls.c_bool_array
    has(output, refs, size, bit_index)
    return np.array(output[0:size], dtype=bool)
  return ret_has 


def create_float_group_has(refs, field, has, metadata):
  """
  Returns a callable which accepts 0 arguments
  """
  bit_index = metadata.float_field_bitfield_offset_ + int(field)
  return __create_has_func(refs, bit_index, has)


def create_high_res_angle_group_has(refs, field, has, metadata):
  """
  Returns a callable which accepts 0 arguments
  """
  bit_index = metadata.high_res_angle_field_bitfield_offset_ + int(field)
  return __create_has_func(refs, bit_index, has)


def create_vector3f_group_has(refs, field, has, metadata):
  """
  Returns a callable which accepts 0 arguments
  """
  bit_index = metadata.vector3f_field_bitfield_offset_ + int(field)
  return __create_has_func(refs, bit_index, has)


def create_quaternionf_group_has(refs, field, has, metadata):
  """
  Returns a callable which accepts 0 arguments
  """
  bit_index = metadata.quaternionf_field_bitfield_offset_ + int(field)
  return __create_has_func(refs, bit_index, has)


def create_uint64_group_has(refs, field, has, metadata):
  """
  Returns a callable which accepts 0 arguments
  """
  bit_index = metadata.uint64_field_bitfield_offset_ + int(field)
  return __create_has_func(refs, bit_index, has)


def create_enum_group_has(refs, field, has, metadata):
  """
  Returns a callable which accepts 0 arguments
  """
  bit_index = metadata.enum_field_bitfield_offset_ + int(field)
  return __create_has_func(refs, bit_index, has)


def create_bool_group_has(refs, field, has, metadata):
  """
  Returns a callable which accepts 0 arguments
  """
  bit_index = metadata.bool_field_bitfield_offset_ + int(field)
  return __create_has_func(refs, bit_index, has)


def create_numbered_float_group_has(refs, field, has, metadata):
  """
  Returns a callable which accepts 1 argument
  """
  relative_offset = int(metadata.numbered_float_relative_offsets_[int(field)])
  bit_index = metadata.numbered_float_field_bitfield_offset_ + relative_offset
  size = len(refs)

  def ret_has(number):
    _tls.ensure_capacity(size)
    output = _tls.c_bool_array
    has(output, refs, size, bit_index + number)
    return np.array(output[0:size], dtype=bool)
  return ret_has


def create_io_group_has(refs, has):
  """
  Returns a callable which accepts 2 arguments
  """
  size = len(refs)

  def ret_has(field, number):
    _tls.ensure_capacity(size)
    output = _tls.c_bool_array
    has(output, refs, size, number - 1, field)
    return np.array(output[0:size], dtype=bool)
  return ret_has


def create_led_group_has(refs, field, has, metadata):
  """
  Returns a callable which accepts 0 arguments
  """
  bit_index = metadata.led_field_bitfield_offset_ + int(field)
  return __create_has_func(refs, bit_index, has)


################################################################################
# `getter` creators
################################################################################


def create_numbered_float_group_getter(refs, field, getter):
  """
  Returns a callable which accepts 1 argument
  """
  size = len(refs)
  def ret_getter(number):
    _tls.ensure_capacity(size)
    output = _tls.c_float_array
    getter(output, refs, size, number - 1, field)
    return np.array(output[0:size], dtype=np.float32)
  return ret_getter


def create_io_float_group_getter(refs, getter):
  """
  Returns a callable which accepts 2 arguments
  """
  size = len(refs)
  def ret_getter(field, number):
    _tls.ensure_capacity(size)
    output = _tls.c_float_array
    getter(output, refs, size, number - 1, field)
    return np.array(output[0:size], dtype=np.float32)
  return ret_getter


def create_io_int_group_getter(refs, getter):
  """
  Returns a callable which accepts 2 arguments
  """
  size = len(refs)
  def ret_getter(field, number):
    _tls.ensure_capacity(size)
    output = _tls.c_int64_array
    getter(output, refs, size, number - 1, field)
    return np.array(output[0:size], dtype=np.int64)
  return ret_getter


def create_led_group_getter(refs, field, getter):
  """
  Returns a callable which accepts 0 arguments
  """
  size = len(refs)
  from hebi._internal.graphics import color_from_int
  def ret_getter():
    _tls.ensure_capacity(size)
    output = _tls.c_int32_array
    getter(output, refs, size, field)
    return [color_from_int(val) for val in output[0:size]]
  return ret_getter


################################################################################
# `setter` creators
################################################################################


def create_numbered_float_group_setter(refs, field, setter):
  """
  Returns a callable which accepts 2 arguments
  """
  size = len(refs)
  def ret_setter(number, value):
    if value is None:
      bfr = None
    else:
      _tls.ensure_capacity(size)
      bfr = _tls.c_float_array
      if hasattr(value, '__len__'):
        for i in range(size):
          bfr[i] = value[i]
      else:
        for i in range(size):
          bfr[i] = value

    setter(refs, bfr, size, number - 1, field)
  return ret_setter


def create_io_float_group_setter(refs, setter):
  """
  Returns a callable which accepts 3 arguments
  """
  size = len(refs)
  def ret_setter(field, number, value):
    if value is None:
      bfr = None
    else:
      _tls.ensure_capacity(size)
      bfr = _tls.c_float_array
      if hasattr(value, '__len__'):
        for i in range(size):
          bfr[i] = value[i]
      else:
        for i in range(size):
          bfr[i] = value

    setter(refs, bfr, size, number - 1, field)
  return ret_setter


def create_io_int_group_setter(refs, setter):
  """
  Returns a callable which accepts 3 arguments
  """
  size = len(refs)
  def ret_setter(field, number, value):
    if value is None:
      bfr = None
    else:
      _tls.ensure_capacity(size)
      bfr = _tls.c_int64_array
      if hasattr(value, '__len__'):
        for i in range(size):
          bfr[i] = value[i]
      else:
        for i in range(size):
          bfr[i] = value

    setter(refs, bfr, size, number - 1, field)
  return ret_setter


def create_led_group_setter(refs, field, setter):
  """
  Returns a callable which accepts 1 argument
  """
  size = len(refs)
  def ret_setter(value):
    _tls.ensure_capacity(size)
    setter(refs, value, size, field)
  return ret_setter


def create_numbered_float_single_getter(ref, field, getter):
  def ret_getter(number):
    ret = _tls.c_float
    getter(byref(ref), byref(ret), 1, number, field)
    return ret.value
  return ret_getter


################################################################################
# Classes
################################################################################


class GroupNumberedFloatFieldContainer(WeakReferenceContainer):
  """
  A read only view into a set of numbered float fields
  """

  __slots__ = ['_getter', '_has', '_field']

  def __init__(self, internal, field, getter, has):
    super(GroupNumberedFloatFieldContainer, self).__init__(internal)
    self._field = field
    self._getter = getter
    self._has = has

  def __repr__(self):
    return _numbered_float_repr(self, self._field)

  @property
  def has_float1(self):
    return self._has(1)

  @property
  def has_float2(self):
    return self._has(2)

  @property
  def has_float3(self):
    return self._has(3)

  @property
  def has_float4(self):
    return self._has(4)

  @property
  def has_float5(self):
    return self._has(5)

  @property
  def has_float6(self):
    return self._has(6)

  @property
  def has_float7(self):
    return self._has(7)

  @property
  def has_float8(self):
    return self._has(8)

  @property
  def has_float9(self):
    return self._has(9)

  @property
  def float1(self):
    return self._getter(1)

  @property
  def float2(self):
    return self._getter(2)

  @property
  def float3(self):
    return self._getter(3)

  @property
  def float4(self):
    return self._getter(4)

  @property
  def float5(self):
    return self._getter(5)

  @property
  def float6(self):
    return self._getter(6)

  @property
  def float7(self):
    return self._getter(7)

  @property
  def float8(self):
    return self._getter(8)

  @property
  def float9(self):
    return self._getter(9)


class MutableGroupNumberedFloatFieldContainer(WeakReferenceContainer):
  """
  A mutable view into a set of numbered float fields
  """

  __slots__ = ['_getter', '_has', '_setter', '_field']

  def __init__(self, internal, field, getter, has, setter):
    super(MutableGroupNumberedFloatFieldContainer, self).__init__(internal)
    self._field = field
    self._getter = getter
    self._has = has
    self._setter = setter

  def __repr__(self):
    return _numbered_float_repr(self, self._field)

  @property
  def has_float1(self):
    return self._has(1)

  @property
  def has_float2(self):
    return self._has(2)

  @property
  def has_float3(self):
    return self._has(3)

  @property
  def has_float4(self):
    return self._has(4)

  @property
  def has_float5(self):
    return self._has(5)

  @property
  def has_float6(self):
    return self._has(6)

  @property
  def has_float7(self):
    return self._has(7)

  @property
  def has_float8(self):
    return self._has(8)

  @property
  def has_float9(self):
    return self._has(9)

  @property
  def float1(self):
    return self._getter(1)

  @property
  def float2(self):
    return self._getter(2)

  @property
  def float3(self):
    return self._getter(3)

  @property
  def float4(self):
    return self._getter(4)

  @property
  def float5(self):
    return self._getter(5)

  @property
  def float6(self):
    return self._getter(6)

  @property
  def float7(self):
    return self._getter(7)

  @property
  def float8(self):
    return self._getter(8)

  @property
  def float9(self):
    return self._getter(9)

  @float1.setter
  def float1(self, value):
    self._setter(1, value)

  @float2.setter
  def float2(self, value):
    self._setter(2, value)

  @float3.setter
  def float3(self, value):
    self._setter(3, value)

  @float4.setter
  def float4(self, value):
    self._setter(4, value)

  @float5.setter
  def float5(self, value):
    self._setter(5, value)

  @float6.setter
  def float6(self, value):
    self._setter(6, value)

  @float7.setter
  def float7(self, value):
    self._setter(7, value)

  @float8.setter
  def float8(self, value):
    self._setter(8, value)

  @float9.setter
  def float9(self, value):
    self._setter(9, value)


class GroupMessageIoFieldBankContainer(WeakReferenceContainer):
  """
  Represents a read only IO bank
  """

  __slots__ = ['_bank', '_bank_readable']

  def __init__(self, bank, bank_readable, io_field_container):
    super(GroupMessageIoFieldBankContainer, self).__init__(io_field_container)
    self._bank = bank
    self._bank_readable = bank_readable.strip().upper()

  def __repr__(self):
    return _io_bank_repr(self, self._bank, self._bank_readable)

  def has_int(self, pin_number):
    """
    Note: `pin_number` indexing starts at `1`
    """
    container = self._get_ref()
    return container.has_int(self._bank, pin_number)

  def has_float(self, pin_number):
    """
    Note: `pin_number` indexing starts at `1`
    """
    container = self._get_ref()
    return container.has_float(self._bank, pin_number)

  def get_int(self, pin_number):
    """
    Note: `pin_number` indexing starts at `1`
    """
    container = self._get_ref()
    return container.get_int(self._bank, pin_number)

  def get_float(self, pin_number):
    """
    Note: `pin_number` indexing starts at `1`
    """
    container = self._get_ref()
    return container.get_float(self._bank, pin_number)


class MutableGroupMessageIoFieldBankContainer(GroupMessageIoFieldBankContainer):
  """
  Represents a mutable IO Bank
  """

  def __init__(self, bank, bank_readable, io_field_container):
    super(MutableGroupMessageIoFieldBankContainer, self).__init__(bank, bank_readable, io_field_container)

  def set_int(self, pin_number, value):
    """
    Note: `pin_number` indexing starts at `1`
    """
    container = self._get_ref()
    return container.set_int(self._bank, pin_number, value)

  def set_float(self, pin_number, value):
    """
    Note: `pin_number` indexing starts at `1`
    """
    container = self._get_ref()
    return container.set_float(self._bank, pin_number, value)


class GroupMessageIoFieldContainer(WeakReferenceContainer):
  """
  Represents a read only view into IO banks
  """

  __slots__ = ['_a', '_b', '_c', '_d', '_e', '_f', '_getter_int', '_getter_float', '_has_int', '_has_float', '__weakref__']

  def __init__(self, group_message, getter_int, getter_float, has_int_func, has_float_func, enum_type, container_type=GroupMessageIoFieldBankContainer):
    super(GroupMessageIoFieldContainer, self).__init__(group_message)

    self._getter_int = getter_int
    self._getter_float = getter_float
    self._has_int = has_int_func
    self._has_float = has_float_func

    bank_a = enum_type.get_enum_value_by_int(0)
    bank_b = enum_type.get_enum_value_by_int(1)
    bank_c = enum_type.get_enum_value_by_int(2)
    bank_d = enum_type.get_enum_value_by_int(3)
    bank_e = enum_type.get_enum_value_by_int(4)
    bank_f = enum_type.get_enum_value_by_int(5)

    self._a = container_type(bank_a, 'a', self)
    self._b = container_type(bank_b, 'b', self)
    self._c = container_type(bank_c, 'c', self)
    self._d = container_type(bank_d, 'd', self)
    self._e = container_type(bank_e, 'e', self)
    self._f = container_type(bank_f, 'f', self)

  def __repr__(self):
    return _io_repr(self)

  def has_int(self, bank, pin_number):
    """
    Note: `pin_number` indexing starts at `1`
    """
    if pin_number < 1:
      raise ValueError("pin_number must be greater than 0")
    return self._has_int(bank, pin_number)

  def has_float(self, bank, pin_number):
    """
    Note: `pin_number` indexing starts at `1`
    """
    if pin_number < 1:
      raise ValueError("pin_number must be greater than 0")
    return self._has_float(bank, pin_number)

  def get_int(self, bank, pin_number):
    """
    Note: `pin_number` indexing starts at `1`
    """
    if pin_number < 1:
      raise ValueError("pin_number must be greater than 0")
    return self._getter_int(bank, pin_number)

  def get_float(self, bank, pin_number):
    """
    Note: `pin_number` indexing starts at `1`
    """
    if pin_number < 1:
      raise ValueError("pin_number must be greater than 0")
    return self._getter_float(bank, pin_number)

  @property
  def a(self):
    return self._a

  @property
  def b(self):
    return self._b

  @property
  def c(self):
    return self._c

  @property
  def d(self):
    return self._d

  @property
  def e(self):
    return self._e

  @property
  def f(self):
    return self._f


class MutableGroupMessageIoFieldContainer(GroupMessageIoFieldContainer):
  """
  Represents a mutable view into IO banks
  """

  __slots__ = ['_setter_int', '_setter_float']

  def __init__(self, group_message, getter_int, getter_float, has_int_func, has_float_func, setter_int, setter_float, enum_type):
    super(MutableGroupMessageIoFieldContainer, self).__init__(group_message, getter_int, getter_float, has_int_func, has_float_func, enum_type, MutableGroupMessageIoFieldBankContainer)

    self._setter_int = setter_int
    self._setter_float = setter_float

  def set_int(self, bank, pin_number, value):
    """
    Note: `pin_number` indexing starts at `1`
    """
    if pin_number < 1:
      raise ValueError("pin_number must be greater than 0")
    self._setter_int(bank, pin_number, value)

  def set_float(self, bank, pin_number, value):
    """
    Note: `pin_number` indexing starts at `1`
    """
    if pin_number < 1:
      raise ValueError("pin_number must be greater than 0")
    self._setter_float(bank, pin_number, value)


################################################################################
# LED Field Containers
################################################################################


def _get_led_values(colors, size):
  _tls.ensure_capacity(size)
  bfr = _tls.c_int32_array

  if type(colors) is str:
    bfr[0:size] = [int(string_to_color(colors)) for _ in range(size)]
  elif type(colors) is int:
    bfr[0:size] = [colors for _ in range(size)]
  elif isinstance(colors, Color):
    bfr[0:size] = [int(colors) for _ in range(size)]
  elif not (hasattr(colors, '__len__')):
    raise ValueError('Cannot broadcast input to array of colors')
  else:
    bfr[0:size] = [int(entry) for entry in colors]
  return bfr


class GroupMessageLEDFieldContainer(WeakReferenceContainer):

  __slots__ = ['_getter', '_field']

  def __init__(self, internal, getter, field):
    super(GroupMessageLEDFieldContainer, self).__init__(internal)
    self._getter = getter
    self._field = field

  def __repr__(self):
    return _led_repr(self, self._field)

  @property
  def color(self):
    return self._getter()


class MutableGroupMessageLEDFieldContainer(GroupMessageLEDFieldContainer):
  __slots__ = ['_setter']

  def __init__(self, internal, getter, setter, field):
    super(MutableGroupMessageLEDFieldContainer, self).__init__(internal, getter, field)
    self._setter = setter

  def clear(self):
    """
    Clears all LEDs
    """
    messages = self._get_ref()
    self._setter(None)

  def __set_colors(self, colors):
    messages = self._get_ref()
    self._setter(_get_led_values(colors, messages.size))

  @property
  def color(self):
    return super(MutableGroupMessageLEDFieldContainer, self).color

  @color.setter
  def color(self, value):
    if value is None:
      self.clear()
    self.__set_colors(value)


################################################################################
# Helper Functions
################################################################################


def __check_broadcastable(group_type, field, value):
  if group_type.size > 1:
    if not field.allow_broadcast:
      raise ValueError('Cannot broadcast scalar value \'{}\' '.format(value) +
                       'to the field \'{0}\' '.format(field.name) +
                       'in all modules of the group.' +
                       '\nReason: {}'.format(field.not_broadcastable_reason))


def __check_broadcastable_field(field, value):
  if not field.allow_broadcast:
    raise ValueError('Cannot broadcast scalar value \'{}\' '.format(value) +\
                     'to the field \'{}\' '.format(field.name) +\
                     'in all modules of the group.' +\
                     '\nReason: {}'.format(field.not_broadcastable_reason))


def _do_broadcast(group_message, field, value, dtype):
  if hasattr(value, '__len__') and type(value) is not str:
    __assert_same_length(group_message, value)
    if dtype is str:
      return value
    else:
      return np.array(value, dtype=dtype)
  else: # Is scalar; each module in group will be set to this
    __check_broadcastable(group_message, field, value)
    return np.array([value] * group_message.size, dtype=dtype)


def __assert_same_length(group_message, b):
  if group_message.size != len(b):
    raise ValueError('Input array must have same size as number of modules in group message')


################################################################################
# TLS for accessors and mutators
################################################################################


import threading


class MessagesTLS_Holder:

  __slots__ = [
                # Scalars
                '_c_bool', '_c_int32', '_c_int64', '_c_uint64', '_c_size_t', '_c_float', '_c_double',
                '_c_vector3f', '_c_quaternionf', '_c_null_str', '_c_str', '_c_high_res_angle',
                "_array_size",
                # Arrays
                "_c_bool_array", "_c_int32_array", "_c_int64_array", "_c_uint64_array", "_c_size_t_array",
                "_c_float_array", "_c_double_array", "_c_vector3f_array", "_c_quaternionf_array",
                "_c_high_res_angle_array"
              ]

  def _grow_arrays(self, size):
    if size > self._array_size:
      self._c_bool_array = (ctypes.c_bool * size)()
      self._c_int32_array = (ctypes.c_int32 * size)()
      self._c_int64_array = (ctypes.c_int64 * size)()
      self._c_uint64_array = (ctypes.c_uint64 * size)()
      self._c_size_t_array = (ctypes.c_size_t * size)()
      self._c_float_array = (ctypes.c_float * size)()
      self._c_double_array = (ctypes.c_double * size)()
      self._c_vector3f_array = (HebiVector3f * size)()
      self._c_quaternionf_array = (HebiQuaternionf * size)()
      self._c_high_res_angle_array = (HebiHighResAngleStruct * size)()

  def __init__(self):
    self._c_bool = ctypes.c_bool(0)
    self._c_int32 = ctypes.c_int32(0)
    self._c_int64 = ctypes.c_int64(0)
    self._c_uint64 = ctypes.c_uint64(0)
    self._c_size_t = ctypes.c_size_t(0)
    self._c_float = ctypes.c_float(0)
    self._c_double = ctypes.c_double(0)
    self._c_vector3f = HebiVector3f()
    self._c_quaternionf = HebiQuaternionf()
    self._c_high_res_angle = HebiHighResAngleStruct()
    self._c_null_str = ctypes.c_char_p(None)
    self._c_str = create_str(512)

    self._array_size = 0
    self._grow_arrays(6)


class MessagesTLS(threading.local):
  def __init__(self):
    super(MessagesTLS, self).__init__()
    self._holder = MessagesTLS_Holder()

  def ensure_capacity(self, size):
    self._holder._grow_arrays(size)

  @property
  def c_bool(self):
    return self._holder._c_bool

  @property
  def c_int32(self):
    return self._holder._c_int32

  @property
  def c_int64(self):
    return self._holder._c_int64

  @property
  def c_uint64(self):
    return self._holder._c_uint64

  @property
  def c_size_t(self):
    return self._holder._c_size_t

  @property
  def c_float(self):
    return self._holder._c_float

  @property
  def c_double(self):
    return self._holder._c_double

  @property
  def c_vector3f(self):
    return self._holder._c_vector3f

  @property
  def c_quaternionf(self):
    return self._holder._c_quaternionf

  @property
  def c_null_str(self):
    return self._holder._c_null_str

  @property
  def c_str(self):
    return self._holder._c_str

  @property
  def c_high_res_angle(self):
    return self._holder._c_high_res_angle

  @property
  def c_bool_array(self):
    return self._holder._c_bool_array

  @property
  def c_int32_array(self):
    return self._holder._c_int32_array

  @property
  def c_int64_array(self):
    return self._holder._c_int64_array

  @property
  def c_uint64_array(self):
    return self._holder._c_uint64_array

  @property
  def c_size_t_array(self):
    return self._holder._c_size_t_array

  @property
  def c_float_array(self):
    return self._holder._c_float_array

  @property
  def c_double_array(self):
    return self._holder._c_double_array

  @property
  def c_vector3f_array(self):
    return self._holder._c_vector3f_array

  @property
  def c_quaternionf_array(self):
    return self._holder._c_quaternionf_array

  @property
  def c_high_res_angle_array(self):
    return self._holder._c_high_res_angle_array


_tls = MessagesTLS()


################################################################################
# Accessors
################################################################################


def __get_flag_group(refs, field, getter):
  size = len(refs)
  _tls.ensure_capacity(size)
  bfr = _tls.c_bool_array
  getter(bfr, refs, size, field)
  return np.array(bfr[0:size], dtype=np.bool)


def __get_bool_group(refs, field, getter):
  size = len(refs)
  _tls.ensure_capacity(size)
  bfr = _tls.c_bool_array
  getter(bfr, refs, size, field)
  return np.array(bfr[0:size], dtype=np.bool)


def __get_enum_group(refs, field, getter):
  size = len(refs)
  _tls.ensure_capacity(size)
  bfr = _tls.c_int32_array
  getter(bfr, refs, size, field)
  return np.array(bfr[0:size], dtype=np.int32)


def __get_float_group(refs, field, getter):
  size = len(refs)
  _tls.ensure_capacity(size)
  bfr = _tls.c_float_array
  getter(bfr, refs, size, field)
  return np.array(bfr[0:size], dtype=np.float32)


def __get_highresangle_group(refs, field, getter):
  size = len(refs)
  _tls.ensure_capacity(size)
  bfr = _tls.c_double_array
  getter(bfr, refs, size, field)
  return np.array(bfr[0:size], dtype=np.float64)


def __get_vector3f_group(refs, field, getter):
  size = len(refs)
  _tls.ensure_capacity(size)
  bfr = _tls.c_vector3f_array
  getter(bfr, refs, size, field)
  return _ctypes_to_ndarray(cast_to_float_ptr(bfr), (size, 3))


def __get_quaternionf_group(refs, field, getter):
  size = len(refs)
  _tls.ensure_capacity(size)
  bfr = _tls.c_quaternionf_array
  getter(bfr, refs, size, field)
  return _ctypes_to_ndarray(cast_to_float_ptr(bfr), (size, 4))


def __get_uint64_group(refs, field, getter):
  size = len(refs)
  _tls.ensure_capacity(size)
  bfr = _tls.c_uint64_array
  getter(bfr, refs, size, field)
  return np.array(bfr[0:size], dtype=np.uint64)


def __get_scalar_field_single(message, field, ret, getter):
  getter(byref(ret), message, 1, field)
  return ret.value


def __get_vector3f_single(message, field, ret, getter):
  ret = byref(ret)
  getter(ret, message, 1, field)
  #return ret.value
  return _ctypes_to_ndarray(cast_to_float_ptr(ret), (3,))


def __get_quaternionf_single(message, field, ret, getter):
  ret = byref(ret)
  getter(ret, message, 1, field)
  #return ret.value
  return _ctypes_to_ndarray(cast_to_float_ptr(ret), (4,))


def __get_string_group(message_list, field, output, getter):
  alloc_size_c = _tls.c_size_t
  alloc_size = 0
  null_str = _tls.c_null_str

  for i, message in enumerate(message_list.modules):
    res = getter(message, field, null_str, byref(alloc_size_c))
    alloc_size = max(alloc_size, alloc_size_c.value + 1)

  if alloc_size > len(_tls.c_str):
    string_buffer = create_str(alloc_size)
  else:
    string_buffer = _tls.c_str

  for i, message in enumerate(message_list.modules):
    alloc_size_c.value = alloc_size
    if getter(message, field, string_buffer, byref(alloc_size_c)) == StatusSuccess:
      output[i] = decode_str(string_buffer.value)
    else:
      output[i] = None
  return output


def __get_string_single(message, field, getter):
  alloc_size_c = _tls.c_size_t
  null_str = _tls.c_null_str

  getter(message, field, null_str, byref(alloc_size_c))
  alloc_size = alloc_size_c.value + 1

  if alloc_size > len(_tls.c_str):
    string_buffer = create_str(alloc_size)
  else:
    string_buffer = _tls.c_str

  alloc_size_c.value = alloc_size
  ret = None
  if getter(message, field, string_buffer, byref(alloc_size_c)) == StatusSuccess:
    ret = decode_str(string_buffer.value)
  return ret


################################################################################
# Mutators
################################################################################


def __set_flag_group(refs, field, value, setter):
  size = len(refs)
  _tls.ensure_capacity(size)
  if value is None:
    bfr = None
  else:
    bfr = _tls.c_bool_array
    if hasattr(value, '__len__'):
      for i in range(size):
        bfr[i] = value[i]
    else:
      for i in range(size):
        bfr[i] = value
  setter(refs, bfr, size, field)


def __set_bool_group(refs, field, value, setter):
  size = len(refs)
  _tls.ensure_capacity(size)
  if value is None:
    bfr = None
  else:
    bfr = _tls.c_bool_array
    if hasattr(value, '__len__'):
      for i in range(size):
        bfr[i] = value[i]
    else:
      for i in range(size):
        bfr[i] = value
  setter(refs, bfr, size, field)


def __set_enum_group(refs, field, value, setter):
  size = len(refs)
  _tls.ensure_capacity(size)
  if value is None:
    bfr = None
  else:
    bfr = _tls.c_int32_array
    if hasattr(value, '__len__'):
      for i in range(size):
        bfr[i] = value[i]
    else:
      for i in range(size):
        bfr[i] = value
  setter(refs, bfr, size, field)


def __set_float_group(refs, field, value, setter):
  size = len(refs)
  _tls.ensure_capacity(size)
  if value is None:
    bfr = None
  else:
    bfr = _tls.c_float_array
    if hasattr(value, '__len__'):
      for i in range(size):
        bfr[i] = value[i]
    else:
      for i in range(size):
        bfr[i] = value
  setter(refs, bfr, size, field)


def __set_highresangle_group(refs, field, value, setter):
  size = len(refs)
  _tls.ensure_capacity(size)
  if value is None:
    bfr = None
  else:
    bfr = _tls.c_double_array
    if hasattr(value, '__len__'):
      for i in range(size):
        bfr[i] = value[i]
    else:
      for i in range(size):
        bfr[i] = value
  setter(refs, bfr, size, field)


def __set_vector3f_group(refs, field, value, setter):
  size = len(refs)
  _tls.ensure_capacity(size)
  if value is None:
    bfr = None
  else:
    bfr = _tls.c_vector3f_array
    if hasattr(value, '__len__'):
      for i in range(size):
        bfr[i] = value[i]
    else:
      for i in range(size):
        bfr[i] = value
  setter(refs, bfr, size, field)


def __set_quaternionf_group(refs, field, value, setter):
  size = len(refs)
  _tls.ensure_capacity(size)
  if value is None:
    bfr = None
  else:
    bfr = _tls.c_quaternionf_array
    if hasattr(value, '__len__'):
      for i in range(size):
        bfr[i] = value[i]
    else:
      for i in range(size):
        bfr[i] = value
  setter(refs, bfr, size, field)


def __set_uint64_group(refs, field, value, setter):
  size = len(refs)
  _tls.ensure_capacity(size)
  if value is None:
    bfr = None
  else:
    bfr = _tls.c_uint64_array
    if hasattr(value, '__len__'):
      for i in range(size):
        bfr[i] = value[i]
    else:
      for i in range(size):
        bfr[i] = value
  setter(refs, bfr, size, field)


def __set_field_single(ref, field, value, value_ctype, setter):
  if value is not None:
    value_ctype.value = value
    setter(byref(ref), byref(value_ctype), 1, field)
  else:
    setter(byref(ref), None, 1, field)


def __set_string_group(message_list, field, value, setter):
  alloc_size_c = _tls.c_size_t
  if value is not None:
    value = _do_broadcast(message_list, field, value, str)

    for i, message in enumerate(message_list.modules):
      val = value[i]
      alloc_size = len(val) + 1
      # TODO: use tls string buffer and copy val into it instead
      string_buffer = create_str(val, size=alloc_size)
      alloc_size_c.value = alloc_size
      setter(message, field, string_buffer, byref(alloc_size_c))
  else:
    for i, message in enumerate(message_list.modules):
      setter(message, field, None, None)


def __set_string_single(message, field, value, setter):
  if value is not None:
    alloc_size_c = _tls.c_size_t
    alloc_size = len(value) + 1
    # TODO: use tls string buffer and copy val into it instead
    string_buffer = create_str(value, size=alloc_size)
    alloc_size_c.value = alloc_size
    setter(message, field, string_buffer, byref(alloc_size_c))
  else:
    setter(message, field, None, None)

################################################################################
# Command
################################################################################


get_command_flag = lambda msg, field: __get_scalar_field_single(msg, field, _tls.c_bool, api.hwCommandGetFlag)
get_command_bool = lambda msg, field: __get_scalar_field_single(msg, field, _tls.c_bool, api.hwCommandGetBool)
get_command_enum = lambda msg, field: __get_scalar_field_single(msg, field, _tls.c_int32, api.hwCommandGetEnum)
get_command_float = lambda msg, field: __get_scalar_field_single(msg, field, _tls.c_float, api.hwCommandGetFloat)
get_command_highresangle = lambda msg, field: __get_scalar_field_single(msg, field, _tls.c_double, api.hwCommandGetHighResAngle)
get_command_string = lambda msg, field: __get_string_single(msg, field, api.hebiCommandGetString)

set_command_flag = lambda msg, field, value: __set_field_single(msg, field, value, _tls.c_bool, api.hwCommandSetFlag)
set_command_bool = lambda msg, field, value: __set_field_single(msg, field, value, _tls.c_bool, api.hwCommandSetBool)
set_command_enum = lambda msg, field, value: __set_field_single(msg, field, value, _tls.c_int32, api.hwCommandSetEnum)
set_command_float = lambda msg, field, value: __set_field_single(msg, field, value, _tls.c_float, api.hwCommandSetFloat)
set_command_highresangle = lambda msg, field, value: __set_field_single(msg, field, value, _tls.c_double, api.hwCommandSetHighResAngle)
set_command_string = lambda msg, field, value: __set_string_single(msg, field, value, api.hebiCommandSetString)

get_group_command_flag = lambda msg, field: __get_flag_group(msg, field, api.hwCommandGetFlag)
get_group_command_bool = lambda msg, field: __get_bool_group(msg, field, api.hwCommandGetBool)
get_group_command_enum = lambda msg, field: __get_enum_group(msg, field, api.hwCommandGetEnum)
get_group_command_float = lambda msg, field: __get_float_group(msg, field, api.hwCommandGetFloat)
get_group_command_highresangle = lambda msg, field: __get_highresangle_group(msg, field, api.hwCommandGetHighResAngle)
get_group_command_string = lambda msg, field, output: __get_string_group(msg, field, output, api.hebiCommandGetString)

set_group_command_flag = lambda msg, field, value: __set_flag_group(msg, field, value, api.hwCommandSetFlag)
set_group_command_bool = lambda msg, field, value: __set_bool_group(msg, field, value, api.hwCommandSetBool)
set_group_command_enum = lambda msg, field, value: __set_enum_group(msg, field, value, api.hwCommandSetEnum)
set_group_command_float = lambda msg, field, value: __set_float_group(msg, field, value, api.hwCommandSetFloat)
set_group_command_highresangle = lambda msg, field, value: __set_highresangle_group(msg, field, value, api.hwCommandSetHighResAngle)
set_group_command_string = lambda msg, field, value: __set_string_group(msg, field, value, api.hebiCommandSetString)


################################################################################
# Feedback
################################################################################


get_feedback_vector3f = lambda msg, field: __get_vector3f_single(msg, field, _tls.c_vector3f, api.hwFeedbackGetVector3f)
get_feedback_quaternionf = lambda msg, field: __get_quaternionf_single(msg, field, _tls.c_quaternionf, api.hwFeedbackGetQuaternionf)
get_feedback_uint64 = lambda msg, field: __get_scalar_field_single(msg, field, _tls.c_uint64, api.hwFeedbackGetUInt64)
get_feedback_enum = lambda msg, field: __get_scalar_field_single(msg, field, _tls.c_int, api.hwFeedbackGetEnum)
get_feedback_float = lambda msg, field: __get_scalar_field_single(msg, field, _tls.c_float, api.hwFeedbackGetFloat)
get_feedback_highresangle = lambda msg, field: __get_scalar_field_single(msg, field, _tls.c_double, api.hwFeedbackGetHighResAngle)

get_group_feedback_vector3f = lambda msg, field: __get_vector3f_group(msg, field, api.hwFeedbackGetVector3f)
get_group_feedback_quaternionf = lambda msg, field: __get_quaternionf_group(msg, field, api.hwFeedbackGetQuaternionf)
get_group_feedback_uint64 = lambda msg, field: __get_uint64_group(msg, field, api.hwFeedbackGetUInt64)
get_group_feedback_enum = lambda msg, field: __get_enum_group(msg, field, api.hwFeedbackGetEnum)
get_group_feedback_float = lambda msg, field: __get_float_group(msg, field, api.hwFeedbackGetFloat)
get_group_feedback_highresangle = lambda msg, field: __get_highresangle_group(msg, field, api.hwFeedbackGetHighResAngle)


def get_group_feedback_float_into(refs, field, output):
  size = len(refs)
  bfr = to_float_ptr(output)
  api.hwFeedbackGetFloat(bfr, refs, size, field)


def get_group_feedback_highresangle_into(refs, field, output):
  size = len(refs)
  bfr = to_double_ptr(output)
  api.hwFeedbackGetHighResAngle(bfr, refs, size, field)


################################################################################
# Info
################################################################################


get_info_flag = lambda msg, field: __get_scalar_field_single(msg, field, _tls.c_bool, api.hwInfoGetFlag)
get_info_bool = lambda msg, field: __get_scalar_field_single(msg, field, _tls.c_bool, api.hwInfoGetBool)
get_info_enum = lambda msg, field: __get_scalar_field_single(msg, field, _tls.c_int, api.hwInfoGetEnum)
get_info_float = lambda msg, field: __get_scalar_field_single(msg, field, _tls.c_float, api.hwInfoGetFloat)
get_info_highresangle = lambda msg, field: __get_scalar_field_single(msg, field, _tls.c_double, api.hwInfoGetHighResAngle)
get_info_string = lambda msg, field: __get_string_single(msg, field, api.hebiInfoGetString)


get_group_info_flag = lambda msg, field: __get_flag_group(msg, field, api.hwInfoGetFlag)
get_group_info_enum = lambda msg, field: __get_enum_group(msg, field, api.hwInfoGetEnum)
get_group_info_bool = lambda msg, field: __get_bool_group(msg, field, api.hwInfoGetBool)
get_group_info_float = lambda msg, field: __get_float_group(msg, field, api.hwInfoGetFloat)
get_group_info_highresangle = lambda msg, field: __get_highresangle_group(msg, field, api.hwInfoGetHighResAngle)
get_group_info_string = lambda msg, field, output: __get_string_group(msg, field, output, api.hebiInfoGetString)


################################################################################
# Parsers
################################################################################


def __map_input_setter_delegate(group_message, values, setter, setter_field):
  mapped_values = [None] * group_message.size
  str_map = setter_field.substrates

  try:
    if type(values) is str:
      val = str_map[values.lower()]
      for i in range(0, group_message.size):
        mapped_values[i] = val
    else:
      for i, entry in enumerate(values):
        mapped_values[i] = str_map[entry.lower()]
  except KeyError as key:
    print('{} is not a valid string parameter.'.format(key))
    print('Valid string parameters: {0}'.format("'" + "', '".join(setter_field.substrates.keys()) + "'"))
    raise key
  setter(group_message, setter_field, mapped_values)


def setter_input_parser_delegate(group_message, values, setter, setter_field):
  """
  Maps strings (case insensitive) to a non-string type.
  Only used for fields which are not of type string

  This function assumes that `setter_field` has an attribute called `substrates`
  which returns a dictionary with :type str: keys and values of the type which
  the function `setter` expects
  """
  if not hasattr(setter_field, 'substrates'):
    raise RuntimeError('Field {} has no substrates map field'.format(setter_field))
  if hasattr(values, '__len__') and type(values) is not str:  # Is "array-like" according to numpy
    for entry in values:
      if type(entry) is not str:
        # By default, delegate to regular `setter` routine
        setter(group_message, setter_field, values)
        return
    __map_input_setter_delegate(group_message, values, setter, setter_field)
  elif type(values) == str:
    __map_input_setter_delegate(group_message, values, setter, setter_field)
  else:
    setter(group_message, setter_field, values)



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
#
#  HEBI Core python API - Copyright 2018 HEBI Robotics
#  See https://hebi.us/softwarelicense for license details
#
# -----------------------------------------------------------------------------


import numpy as np
from math import pi
from threading import local
from .utils import CaseInvariantString, intern_string

neg_half_pi = pi*-0.5


################################################################################
# Transform Functions
################################################################################


def set_translate(matrix, x, y, z):
  matrix[0, 3] = x
  matrix[1, 3] = y
  matrix[2, 3] = z


def set_rotate_x(matrix, radians):
  from math import cos, sin
  c_r = cos(radians)
  s_r = sin(radians)
  matrix[0, 0] = 1.0
  matrix[0, 1] = 0.0
  matrix[0, 2] = 0.0
  matrix[1, 0] = 0.0
  matrix[1, 1] = c_r
  matrix[1, 2] = -s_r
  matrix[2, 0] = 0.0
  matrix[2, 1] = s_r
  matrix[2, 2] = c_r


def set_sphere_inertia(inertia, mass, radius):
  inertia[0:3] = 0.4 * mass * radius * radius
  inertia[3:6] = 0.0


def set_rod_x_axis_inertia(inertia, mass, length):
  inertia[1:3] = mass * length * length * 0.083333333333333333
  inertia[3:6] = inertia[0] = 0.0


################################################################################
# Descriptor classes returned from parser functions
################################################################################


class ActuatorDesc(object):

  __slots__ = ['_com', '_com_xyz', '_inertia', '_input_to_axis', '_input_to_axis_xyz', '_mass', '_name']

  def __init__(self, name, mass, inertia, com_xyz, input_to_axis_xyz):
    self._name = name
    self._mass = mass
    self._inertia = np.array(inertia, dtype=np.float64)
    self._com_xyz = com_xyz
    self._input_to_axis_xyz = input_to_axis_xyz
    self._com = np.identity(4, np.float64)
    self._input_to_axis = np.identity(4, np.float64)

    v = com_xyz
    set_translate(self._com, v[0], v[1], v[2])
    v = input_to_axis_xyz
    set_translate(self._input_to_axis, v[0], v[1], v[2])

  @property
  def name(self):
    return self._name

  @property
  def mass(self):
    return self._mass

  @property
  def moments_of_inertia(self):
    return self._inertia

  @property
  def com(self):
    return self._com

  @property
  def input_to_axis(self):
    return self._input_to_axis


class BracketDesc(object):

  __slots__ = ['_mass', '_mount', '_name']

  def __init__(self, name, mount, mass):
    self._name = name
    self._mount = mount
    self._mass = mass

  @property
  def name(self):
    return self._name

  @property
  def mount(self):
    return self._mount

  @property
  def mass(self):
      return self._mass


################################################################################
# Descriptor Classes for parsing
################################################################################


class BracketParserMatch(object):

  __slots__ = ['_mount', '_parser']

  def __init__(self, parser, mount):
    self._parser = parser
    self._mount = mount

  @property
  def bracket_name(self):
    return self._parser.name

  @property
  def light(self):
    return 'Light' in self._parser.name

  @property
  def heavy(self):
    return 'Heavy' in self._parser.name

  @property
  def left(self):
    return 'left' in self._mount

  @property
  def right(self):
    return 'right' in self._mount

  @property
  def inside(self):
    return 'inside' in self._mount

  @property
  def outside(self):
    return 'outside' in self._mount

  @property
  def mass(self):
    return self._parser.mass


class BracketParser(object):

  __slots__ = ['_mass', '_mounts', '_name']

  def __init__(self, name, mounts, mass):
    self._name = intern_string(name)
    self._mounts = [intern_string(mount) for mount in mounts]
    self._mass = mass

  def match(self, name, mount):
    if name == self._name:
      mount = intern_string(mount)
      for entry in self._mounts:
        if mount == entry:
          return BracketParserMatch(self, mount)
    return None

  @property
  def mass(self):
    return self._mass

  @property
  def name(self):
    return self._name


################################################################################
# Maps and Lists for HEBI Products
################################################################################


class KinematicsParserTLS(local):
  """
  Used to cache valid strings mapped to kinematics objects
  """

  def __init__(self):
    super(KinematicsParserTLS, self).__init__()
    self.parsed_joint_strs = dict()
    self.parsed_frame_type_strs = dict()
    self.parsed_actuator_strs = dict()
    self.parsed_link_strs = dict()
    self.parsed_bracket_strs = dict()

  def get_joint(self, joint):
    return self.parsed_joint_strs.get(joint)

  def set_joint(self, joint, value):
    self.parsed_joint_strs[joint] = value

  def get_frame_type(self, frame_type):
    return self.parsed_frame_type_strs.get(frame_type)

  def set_frame_type(self, frame_type, value):
    self.parsed_frame_type_strs[frame_type] = value

  def get_actuator(self, actuator):
    return self.parsed_actuator_strs.get(actuator)

  def set_actuator(self, actuator, value):
    self.parsed_actuator_strs[actuator] = value


_parser_tls = KinematicsParserTLS()

# TODO: Maybe move this into a catalogue module?
from .ffi import enums

__joint_types_to_enum = {
  'tx': enums.JointTypeTranslationX, 'x': enums.JointTypeTranslationX,
  'ty': enums.JointTypeTranslationY, 'y': enums.JointTypeTranslationY,
  'tz': enums.JointTypeTranslationZ, 'z': enums.JointTypeTranslationZ,
  'rx': enums.JointTypeRotationX,
  'ry': enums.JointTypeRotationY,
  'rz': enums.JointTypeRotationZ
  }

__enum_to_joint_types = dict((v,k) for k,v in __joint_types_to_enum.items())

__end_effector_types_to_enum = {
  'custom' : enums.EndEffectorTypeCustom,
  'x5' : enums.EndEffectorTypeX5Parallel, 'x8' : enums.EndEffectorTypeX5Parallel,
  'X5' : enums.EndEffectorTypeX5Parallel, 'X8' : enums.EndEffectorTypeX5Parallel,
  'r8' : enums.EndEffectorTypeR8Parallel, 'R8' : enums.EndEffectorTypeR8Parallel,
  }

__enum_to_end_effector_types = dict((v,k) for k,v in __end_effector_types_to_enum.items())

__frame_types_to_enum = {
  'CoM': enums.FrameTypeCenterOfMass,
  'com': enums.FrameTypeCenterOfMass,
  'output': enums.FrameTypeOutput,
  'endeffector': enums.FrameTypeEndEffector, 'EndEffector': enums.FrameTypeEndEffector,
  'input': enums.FrameTypeInput,
  }

__X5_moi = [0.00015, 0.000255, 0.000350, 0.0000341, 0.0000118, 0.00000229]
__X8_moi = [0.000246, 0.000380, 0.000463, 0.0000444, 0.0000266, 0.00000422]
__R8_moi = [0.000246, 0.000380, 0.000463, 0.0000444, 0.0000266, 0.00000422] # TODO

__actuators = {
  'X5-1': ActuatorDesc('X5-1', 0.315, __X5_moi, [0.0142, -0.0031, 0.0165], [0.0, 0.0, 0.03105]),
  'X5-4': ActuatorDesc('X5-4', 0.335, __X5_moi, [0.0142, -0.0031, 0.0165], [0.0, 0.0, 0.03105]),
  'X5-9': ActuatorDesc('X5-9', 0.360, __X5_moi, [0.0142, -0.0031, 0.0165], [0.0, 0.0, 0.03105]),
  'X8-3': ActuatorDesc('X8-3', 0.460, __X8_moi, [-0.0145, -0.0031, 0.0242], [0.0, 0.0, 0.0451]),
  'X8-9': ActuatorDesc('X8-9', 0.480, __X8_moi, [-0.0145, -0.0031, 0.0242], [0.0, 0.0, 0.0451]),
  'X8-16': ActuatorDesc('X8-16', 0.500, __X8_moi, [-0.0145, -0.0031, 0.0242], [0.0, 0.0, 0.0451]),
  'R8-3': ActuatorDesc('R8-3', 0.500, __R8_moi, [-0.0145, -0.0031, 0.0242], [0.0, 0.0, 0.0451]),
  'R8-9': ActuatorDesc('R8-9', 0.500, __R8_moi, [-0.0145, -0.0031, 0.0242], [0.0, 0.0, 0.0451]),
  'R8-16': ActuatorDesc('R8-16', 0.500, __R8_moi, [-0.0145, -0.0031, 0.0242], [0.0, 0.0, 0.0451]),
  }

__actuator_to_enum = {
  CaseInvariantString('X5-1'): enums.ActuatorTypeX5_1,
  CaseInvariantString('X5-4'): enums.ActuatorTypeX5_4,
  CaseInvariantString('X5-9'): enums.ActuatorTypeX5_9,
  CaseInvariantString('X8-3'): enums.ActuatorTypeX8_3,
  CaseInvariantString('X8-9'): enums.ActuatorTypeX8_9,
  CaseInvariantString('X8-16'): enums.ActuatorTypeX8_16,
  CaseInvariantString('R8-3'): enums.ActuatorTypeR8_3,
  CaseInvariantString('R8-9'): enums.ActuatorTypeR8_9,
  CaseInvariantString('R8-16'): enums.ActuatorTypeR8_16,
}

__enum_to_actuator = dict((v, str(k)) for k,v in __actuator_to_enum.items())


def _make_tup(a, b):
  return (CaseInvariantString(a), CaseInvariantString(b))


__actuator_brackets_to_enum = {
  _make_tup('X5-LightBracket', 'left'): enums.BracketTypeX5LightLeft,
  _make_tup('X5-LightBracket', 'right'): enums.BracketTypeX5LightRight,
  _make_tup('X5-HeavyBracket', 'left-inside'): enums.BracketTypeX5HeavyLeftInside,
  _make_tup('X5-HeavyBracket', 'right-inside'): enums.BracketTypeX5HeavyRightInside,
  _make_tup('X5-HeavyBracket', 'left-outside'): enums.BracketTypeX5HeavyLeftOutside,
  _make_tup('X5-HeavyBracket', 'right-outside'): enums.BracketTypeX5HeavyRightOutside,
  _make_tup('R8-LightBracket', 'left'): enums.BracketTypeR8LightLeft,
  _make_tup('R8-LightBracket', 'right'): enums.BracketTypeR8LightRight,
  _make_tup('R8-HeavyBracket', 'left-inside'): enums.BracketTypeR8HeavyLeftInside,
  _make_tup('R8-HeavyBracket', 'right-inside'): enums.BracketTypeR8HeavyRightInside,
  _make_tup('R8-HeavyBracket', 'left-outside'): enums.BracketTypeR8HeavyLeftOutside,
  _make_tup('R8-HeavyBracket', 'right-outside'): enums.BracketTypeR8HeavyRightOutside,
  }

__enum_to_actuator_brackets = dict((v, str(k[0]) + ' ({0})'.format(k[1])) for k,v in __actuator_brackets_to_enum.items())

__actuator_links_to_enum = {
  CaseInvariantString('X5'): enums.LinkTypeX5, CaseInvariantString('X8'): enums.LinkTypeX5,
  CaseInvariantString('R8'): enums.LinkTypeR8,
}

__enum_to_actuator_links = dict((v, str(k)) for k,v in __actuator_links_to_enum.items())

__actuator_links = {
  CaseInvariantString('X5'), CaseInvariantString('X8'),
  CaseInvariantString('R8'),
  }

__brackets = [
  BracketParser('X5-LightBracket',
                ['left', 'right'], 0.1),
  BracketParser('X5-HeavyBracket',
                ['left-inside', 'right-inside',
                 'left-outside', 'right-outside'], 0.215),
  BracketParser('R8-LightBracket',
                ['left', 'right'], 0.1), # FIXME: MASS
  BracketParser('R8-HeavyBracket',
                ['left-inside', 'right-inside',
                 'left-outside', 'right-outside'], 0.215), # FIXME: MASS
]


################################################################################
# Parsing Functions
################################################################################


def __assert_str(value, name='value'):
  if not isinstance(value, str):
    raise TypeError('{} must be a str'.format(name))


def __get_from_dict(value, dict_, in_type):
  ret = dict_.get(value, None)
  if ret is not None:
    return ret
  raise ValueError("{0} is not a valid {1}. ".format(value, in_type) +
                   "Valid types are: {0}".format(
                     "'" + "', '".join([str(entry) for entry in dict_.keys()]) + "'"))


def parse_frame_type(value):
  tls_val = _parser_tls.get_frame_type(value)
  if tls_val is not None:
    return tls_val
  # If not cached:
  try:
    tls_val = __frame_types_to_enum[value]
    _parser_tls.set_frame_type(value, tls_val)
    return tls_val
  except Exception as e:
    __assert_str(value, 'frame type')
    if type(e) is KeyError:
      raise ValueError('{} is not a valid frame type'.format(value))
    else:
      raise e


def parse_joint_type(value):
  tls_val = _parser_tls.get_joint(value)
  if tls_val is not None:
    return tls_val
  # If not cached:
  try:
    fixed_val = value.lower()
    tls_val = __joint_types_to_enum[fixed_val]
    _parser_tls.set_joint(value, tls_val)
    return tls_val
  except Exception as e:
    __assert_str(value, 'joint type')
    if type(e) is KeyError:
      raise ValueError('{} is not a valid joint type'.format(value))
    else:
      raise e


def parse_actuator(value):
  tls_val = _parser_tls.get_actuator(value)
  if tls_val is not None:
    return tls_val
  # If not cached:
  try:
    fixed_val = value.upper()
    tls_val = __actuators[fixed_val]
    _parser_tls.set_frame_type(value, tls_val)
    return tls_val, tls_val.com, tls_val.input_to_axis
  except Exception as e:
    __assert_str(value, 'actuator')
    if type(e) is KeyError:
      raise ValueError('{} is not a valid actuator'.format(value))
    else:
      raise e


def parse_actuator_link(value, extension, twist):
  __assert_str(value)

  value = value.strip().upper()
  if not (value in __actuator_links):
    raise ValueError('{} is not a valid actuator link'.format(value))

  try:
    extension = float(extension)
  except ValueError as v:
    raise ValueError('cannot convert extension={} to a float'.format(extension))
  except TypeError as t:
    raise TypeError('cannot convert extension={} (type {}) to a float'.format(extension, type(extension)))

  try:
    twist = float(twist)
  except ValueError as v:
    raise ValueError('cannot convert twist={} to a float'.format(twist))
  except TypeError as t:
    raise TypeError('cannot convert twist={} (type {}) to a float'.format(twist, type(twist)))

  return extension, twist


def parse_bracket(bracket, mount):
  __assert_str(bracket, 'bracket')
  __assert_str(mount, 'mount')

  bracket = bracket.strip()
  mount = mount.strip().lower()

  for entry in __brackets:
    match = entry.match(bracket, mount)
    if match:
      break
  else:
    # Should never happen, but let's make this lint-free
    match = None

  if not match:
    raise ValueError('bracket={} and mount={} is not a valid bracket'.format(bracket, mount))

  com = np.identity(4, np.float64)
  output = np.identity(4, np.float64)
  mass = match.mass

  if match.light:
    if match.right:
      mult = -1.0
    else:
      mult = 1.0

    set_translate(com, 0.0, mult*0.0215, 0.02)
    set_rotate_x(output, mult*neg_half_pi)
    set_translate(output, 0.0, mult*0.043, 0.04)

  elif match.heavy:
    if match.right:
      lr_mult = -1.0
    else:
      lr_mult = 1.0

    if match.outside:
      y_dist = 0.0375
    else:
      y_dist = -0.0225

    set_translate(com, 0.0, lr_mult * 0.5 * y_dist, 0.0275)
    set_rotate_x(output, lr_mult * neg_half_pi)
    set_translate(output, 0.0, lr_mult * y_dist, 0.055)

  else:
    raise RuntimeError('Unknown bracket type {}'.format(match.bracket_name))

  return com, output, mass


################################################################################
# String to enum
################################################################################


def actuator_str_to_enum(actuator):
  __assert_str(actuator, 'actuator')
  actuator = CaseInvariantString(actuator)
  if actuator not in __actuator_to_enum:
    raise ValueError('Unknown actuator type {}'.format(actuator))
  return __actuator_to_enum[actuator]


def bracket_str_to_enum(bracket, mount):
  __assert_str(bracket, 'bracket')
  __assert_str(mount, 'mount')

  k = _make_tup(bracket, mount)
  if k not in __actuator_brackets_to_enum:
    raise ValueError('Unknown bracket type (bracket: {}, mount: {})'.format(bracket, mount))
  return __actuator_brackets_to_enum[k]


def link_str_to_enum(link):
  __assert_str(link, 'link')
  link = CaseInvariantString(link)

  if link not in __actuator_links_to_enum:
    raise ValueError('Unknown link type {}'.format(link))
  return __actuator_links_to_enum[link]


def end_effector_str_to_enum(end_effector):
  __assert_str(end_effector, 'end_effector')

  if end_effector not in __end_effector_types_to_enum:
    raise ValueError('Unknown end effector type {}'.format(end_effector))
  return __end_effector_types_to_enum[end_effector]


################################################################################
# Enum to string
################################################################################


def _add_pure_ints(d):
  to_add = list()
  for key in d.keys():
    if type(key) is not int:
      to_add.append((int(key), d[key]))
  for entry in to_add:
    d[entry[0]] = entry[1]


_add_pure_ints(__enum_to_actuator)
_add_pure_ints(__enum_to_actuator_brackets)
_add_pure_ints(__enum_to_actuator_links)
_add_pure_ints(__enum_to_joint_types)

actuator_enum_to_str = lambda actuator: __enum_to_actuator[actuator]
bracket_enum_to_str = lambda bracket: __enum_to_actuator_brackets[bracket]
link_enum_to_str = lambda link: __enum_to_actuator_links[link]
joint_enum_to_str = lambda joint: __enum_to_joint_types[joint]
end_effector_enum_to_str = lambda joint: __enum_to_end_effector_types[joint]



import ctypes as _ctypes
import numpy as _np


################################################################################
### Common ctypes classes and values
################################################################################


c_int_p = _ctypes.POINTER(_ctypes.c_int)
c_int32_p = _ctypes.POINTER(_ctypes.c_int32)
c_uint32_p = _ctypes.POINTER(_ctypes.c_uint32)
c_int64_p = _ctypes.POINTER(_ctypes.c_int64)
c_uint64_p = _ctypes.POINTER(_ctypes.c_uint64)
c_float_p = _ctypes.POINTER(_ctypes.c_float)
c_double_p = _ctypes.POINTER(_ctypes.c_double)

NULLPTR = _ctypes.c_void_p(0)


################################################################################
### Common ctypes functions
################################################################################


to_ctypes = _np.ctypeslib.as_ctypes
cast = _ctypes.cast
byref = _ctypes.byref


cast_to_double_ptr = lambda val: cast(val, c_double_p)
cast_to_float_ptr = lambda val: cast(val, c_float_p)


def pointer_offset(ptr, offset):
  """
  Equivalent to `(char*)ptr + offset` or `&((char*)ptr)[offset]`
  """
  return byref(ptr.contents, offset)


def to_double_ptr(val):
  """
  Retrieve a ctypes pointer to the provided numpy object
  """
  return cast(val.__array_interface__['data'][0], c_double_p)


def to_float_ptr(val):
  """
  Retrieve a ctypes pointer to the provided numpy object
  """
  return cast(val.__array_interface__['data'][0], c_float_p)
